import asyncio

import pytest

import aioftp


@pytest.mark.asyncio
async def test_client_without_server(pair_factory, unused_tcp_port_factory):
    f = pair_factory(connected=False, logged=False, do_quit=False)
    async with f as pair:
        pass
    with pytest.raises(OSError):
        await pair.client.connect("127.0.0.1", unused_tcp_port_factory())


@pytest.mark.asyncio
async def test_connection(pair_factory):
    async with pair_factory(connected=True, logged=False, do_quit=False):
        pass


@pytest.mark.asyncio
async def test_quit(pair_factory):
    async with pair_factory(connected=True, logged=False, do_quit=True):
        pass


@pytest.mark.asyncio
async def test_not_implemented(pair_factory, expect_codes_in_exception):
    async with pair_factory() as pair:
        with expect_codes_in_exception("502"):
            await pair.client.command("FOOBAR", "2xx", "1xx")


@pytest.mark.asyncio
async def test_type_success(pair_factory, expect_codes_in_exception):
    async with pair_factory() as pair:
        await pair.client.get_passive_connection("A")


@pytest.mark.asyncio
async def test_extra_pasv_connection(pair_factory):
    async with pair_factory() as pair:
        r, w = await pair.client.get_passive_connection()
        er, ew = await pair.client.get_passive_connection()
        with pytest.raises((ConnectionResetError, BrokenPipeError)):
            while True:
                w.write(b"-" * aioftp.DEFAULT_BLOCK_SIZE)
                await w.drain()


@pytest.mark.parametrize("method", ["epsv", "pasv"])
@pytest.mark.asyncio
async def test_closing_passive_connection(pair_factory, method):
    async with pair_factory(host="127.0.0.1") as pair:
        r, w = await pair.client.get_passive_connection(commands=[method])
        host, port, *_ = w.transport.get_extra_info("peername")
        nr, nw = await asyncio.open_connection(host, port)
        with pytest.raises((ConnectionResetError, BrokenPipeError)):
            while True:
                nw.write(b"-" * aioftp.DEFAULT_BLOCK_SIZE)
                await nw.drain()


@pytest.mark.asyncio
async def test_pasv_connection_ports_not_added(pair_factory):
    async with pair_factory() as pair:
        r, w = await pair.client.get_passive_connection()
        assert pair.server.available_data_ports is None


@pytest.mark.asyncio
async def test_pasv_connection_ports(pair_factory, Server,
                                     unused_tcp_port_factory):
    ports = [unused_tcp_port_factory(), unused_tcp_port_factory()]
    async with pair_factory(None, Server(data_ports=ports)) as pair:
        r, w = await pair.client.get_passive_connection()
        host, port, *_ = w.transport.get_extra_info("peername")
        assert port in ports
        assert pair.server.available_data_ports.qsize() == 1


@pytest.mark.asyncio
async def test_data_ports_remains_empty(pair_factory, Server):
    async with pair_factory(None, Server(data_ports=[])) as pair:
        assert pair.server.available_data_ports.qsize() == 0


@pytest.mark.asyncio
async def test_pasv_connection_port_reused(pair_factory, Server,
                                           unused_tcp_port):
    s = Server(data_ports=[unused_tcp_port])
    async with pair_factory(None, s) as pair:
        r, w = await pair.client.get_passive_connection()
        host, port, *_ = w.transport.get_extra_info("peername")
        assert port == unused_tcp_port
        assert pair.server.available_data_ports.qsize() == 0
        w.close()
        await pair.client.quit()
        pair.client.close()
        assert pair.server.available_data_ports.qsize() == 1
        await pair.client.connect(pair.server.server_host,
                                  pair.server.server_port)
        await pair.client.login()
        r, w = await pair.client.get_passive_connection()
        host, port, *_ = w.transport.get_extra_info("peername")
        assert port == unused_tcp_port
        assert pair.server.available_data_ports.qsize() == 0


@pytest.mark.parametrize("method", ["epsv", "pasv"])
@pytest.mark.asyncio
async def test_pasv_connection_no_free_port(pair_factory, Server,
                                            expect_codes_in_exception,
                                            method):
    s = Server(data_ports=[])
    async with pair_factory(None, s, do_quit=False, host="127.0.0.1") as pair:
        assert pair.server.available_data_ports.qsize() == 0
        with expect_codes_in_exception("421"):
            await pair.client.get_passive_connection(commands=[method])


@pytest.mark.asyncio
async def test_pasv_connection_busy_port(pair_factory, Server,
                                         unused_tcp_port_factory):
    ports = [unused_tcp_port_factory(), unused_tcp_port_factory()]
    async with pair_factory(None, Server(data_ports=ports)) as pair:
        conflicting_server = await asyncio.start_server(
            lambda r, w: w.close(),
            host=pair.server.server_host,
            port=ports[0],
        )
        r, w = await pair.client.get_passive_connection()
        host, port, *_ = w.transport.get_extra_info("peername")
        assert port == ports[1]
        assert pair.server.available_data_ports.qsize() == 1
    conflicting_server.close()
    await conflicting_server.wait_closed()


@pytest.mark.asyncio
async def test_pasv_connection_busy_port2(pair_factory, Server,
                                          unused_tcp_port_factory,
                                          expect_codes_in_exception):
    ports = [unused_tcp_port_factory()]
    s = Server(data_ports=ports)
    async with pair_factory(None, s, do_quit=False) as pair:
        conflicting_server = await asyncio.start_server(
            lambda r, w: w.close(),
            host=pair.server.server_host,
            port=ports[0],
        )
        with expect_codes_in_exception("421"):
            await pair.client.get_passive_connection()
    conflicting_server.close()
    await conflicting_server.wait_closed()


@pytest.mark.asyncio
async def test_server_shutdown(pair_factory):
    async with pair_factory(do_quit=False) as pair:
        await pair.client.list()
        await pair.server.close()
        with pytest.raises(ConnectionResetError):
            await pair.client.list()


@pytest.mark.asyncio
async def test_client_session_context_manager(pair_factory):
    async with pair_factory(connected=False) as pair:
        async with aioftp.ClientSession(*pair.server.address) as client:
            await client.list()


@pytest.mark.asyncio
async def test_long_login_sequence_fail(pair_factory,
                                        expect_codes_in_exception):
    class CustomServer(aioftp.Server):
        async def user(self, connection, rest):
            connection.response("331")
            return True

        async def pass_(self, connection, rest):
            connection.response("332")
            return True

        async def acct(self, connection, rest):
            connection.response("333")
            return True

    factory = pair_factory(logged=False, server_factory=CustomServer,
                           do_quit=False)
    async with factory as pair:
        with expect_codes_in_exception("333"):
            await pair.client.login()


@pytest.mark.asyncio
async def test_bad_sublines_seq(pair_factory, expect_codes_in_exception):
    class CustomServer(aioftp.Server):
        async def write_response(self, stream, code, lines="", list=False):
            import functools
            lines = aioftp.wrap_with_container(lines)
            write = functools.partial(self.write_line, stream)
            *body, tail = lines
            for line in body:
                await write(code + "-" + line)
            await write(str(int(code) + 1) + "-" + tail)
            await write(code + " " + tail)

    factory = pair_factory(connected=False, server_factory=CustomServer)
    async with factory as pair:
        with expect_codes_in_exception("220"):
            await pair.client.connect(pair.server.server_host,
                                      pair.server.server_port)
            await pair.client.login()



import pytest

import aioftp


@pytest.mark.asyncio
async def test_stream_iter_by_line(pair_factory):
    async with pair_factory() as pair:
        await pair.client.make_directory("bar")
        lines = []
        async with pair.client.get_stream("list") as stream:
            async for line in stream.iter_by_line():
                lines.append(line)
        assert len(lines) == 1
        assert b"bar" in lines[0]


@pytest.mark.asyncio
async def test_stream_close_without_finish(pair_factory):
    class CustomException(Exception):
        pass

    def fake_finish(*a, **kw):
        raise Exception("Finished called")

    async with pair_factory() as pair:
        with pytest.raises(CustomException):
            async with pair.client.get_stream():
                raise CustomException()


@pytest.mark.asyncio
async def test_no_server(unused_tcp_port):
    with pytest.raises(ConnectionRefusedError):
        async with aioftp.ClientSession("127.0.0.1", unused_tcp_port):
            pass


@pytest.mark.asyncio
async def test_syst_command(pair_factory):
    async with pair_factory() as pair:
        code, info = await pair.client.command("syst", "215")
        assert info == [" UNIX Type: L8"]



import pathlib
import textwrap
import contextlib

import pytest

import aioftp


async def not_implemented(connection, rest):
    connection.response("502", ":P")
    return True


@pytest.mark.asyncio
async def test_client_fallback_to_list_at_list(pair_factory):
    async with pair_factory() as pair:
        pair.server.mlst = not_implemented
        pair.server.mlsd = not_implemented
        await pair.make_server_files("bar/foo")
        (path, stat), *_ = files = await pair.client.list()
        assert len(files) == 1
        assert path == pathlib.PurePosixPath("bar")
        assert stat["type"] == "dir"
        (path, stat), *_ = files = await pair.client.list("bar")
        assert len(files) == 1
        assert path == pathlib.PurePosixPath("bar/foo")
        assert stat["type"] == "file"


async def implemented_badly(connection, rest):
    assert False, "should not be called"


@pytest.mark.asyncio
async def test_client_list_override(pair_factory):
    async with pair_factory() as pair:
        pair.server.mlsd = implemented_badly
        await pair.client.make_directory("bar")
        (path, stat), *_ = files = await pair.client.list(raw_command="LIST")
        assert len(files) == 1
        assert path == pathlib.PurePosixPath("bar")
        assert stat["type"] == "dir"


@pytest.mark.asyncio
async def test_client_list_override_invalid_raw_command(pair_factory):
    async with pair_factory() as pair:
        with pytest.raises(ValueError):
            await pair.client.list(raw_command="FOO")


def test_client_list_windows():
    test_str = textwrap.dedent("""\
         11/4/2018   9:09 PM  <DIR>         .
         8/10/2018   1:02 PM  <DIR>         ..
         9/23/2018   2:16 PM  <DIR>         bin
        10/16/2018  10:25 PM  <DIR>         Desktop
         11/4/2018   3:31 PM  <DIR>         dow
        10/16/2018   8:21 PM  <DIR>         Downloads
        10/14/2018   5:34 PM  <DIR>         msc
          9/9/2018   9:32 AM  <DIR>         opt
         10/3/2018   2:58 PM    34,359,738,368  win10.img
         6/30/2018   8:36 AM    3,939,237,888  win10.iso
         7/26/2018   1:11 PM           189  win10.sh
        10/29/2018  11:46 AM    34,359,738,368  win7.img
         6/30/2018   8:35 AM    3,319,791,616  win7.iso
        10/29/2018  10:55 AM           219  win7.sh
               6 files           75,978,506,648 bytes
               3 directories     22,198,362,112 bytes free
    """)
    test_str = test_str.strip().split("\n")
    entities = {}
    parse = aioftp.Client(encoding="utf-8").parse_list_line_windows
    for x in test_str:
        with contextlib.suppress(ValueError):
            path, stat = parse(x.encode("utf-8"))
            entities[path] = stat
    dirs = ["bin", "Desktop", "dow", "Downloads", "msc", "opt"]
    files = ["win10.img", "win10.iso", "win10.sh", "win7.img",
             "win7.iso", "win7.sh"]
    assert len(entities) == len(dirs + files)
    for d in dirs:
        p = pathlib.PurePosixPath(d)
        assert p in entities
        assert entities[p]["type"] == "dir"
    for f in files:
        p = pathlib.PurePosixPath(f)
        assert p in entities
        assert entities[p]["type"] == "file"
    with pytest.raises(ValueError):
        parse(b" 10/3/2018   2:58 PM    34,35xxx38,368  win10.img")


@pytest.mark.asyncio
async def test_client_list_override_with_custom(pair_factory, Client):
    meta = {"type": "file", "works": True}

    def parser(b):
        import pickle
        return pickle.loads(bytes.fromhex(b.decode().rstrip("\r\n")))

    async def builder(_, path):
        import pickle
        return pickle.dumps((path, meta)).hex()

    async with pair_factory(Client(parse_list_line_custom=parser)) as pair:
        pair.server.mlst = not_implemented
        pair.server.mlsd = not_implemented
        pair.server.build_list_string = builder
        await pair.client.make_directory("bar")
        (path, stat), *_ = files = await pair.client.list()
        assert len(files) == 1
        assert path == pathlib.PurePosixPath("bar")
        assert stat == meta



from pathlib import PurePosixPath

import pytest

import aioftp


@pytest.mark.asyncio
async def test_remove_single_file(pair_factory):
    async with pair_factory() as pair:
        await pair.make_server_files("foo.txt")
        assert await pair.server_paths_exists("foo.txt")
        await pair.client.remove_file("foo.txt")
        assert await pair.server_paths_exists("foo.txt") is False


@pytest.mark.asyncio
async def test_recursive_remove(pair_factory):
    async with pair_factory() as pair:
        paths = ["foo/bar.txt", "foo/baz.txt", "foo/bar_dir/foo.baz"]
        await pair.make_server_files(*paths)
        await pair.client.remove("foo")
        assert await pair.server_paths_exists(*paths) is False


@pytest.mark.asyncio
async def test_file_download(pair_factory):
    async with pair_factory() as pair:
        await pair.make_server_files("foo", size=1, atom=b"foobar")
        async with pair.client.download_stream("foo") as stream:
            data = await stream.read()
        assert data == b"foobar"


@pytest.mark.asyncio
async def test_file_upload(pair_factory):
    async with pair_factory() as pair:
        async with pair.client.upload_stream("foo") as stream:
            await stream.write(b"foobar")
        async with pair.client.download_stream("foo") as stream:
            data = await stream.read()
        assert data == b"foobar"


@pytest.mark.asyncio
async def test_file_append(pair_factory):
    async with pair_factory() as pair:
        await pair.make_server_files("foo", size=1, atom=b"foobar")
        async with pair.client.append_stream("foo") as stream:
            await stream.write(b"foobar")
        async with pair.client.download_stream("foo") as stream:
            data = await stream.read()
        assert data == b"foobar" * 2


@pytest.mark.asyncio
async def test_upload_folder(pair_factory):
    async with pair_factory() as pair:
        paths = ["foo/bar", "foo/baz"]
        await pair.make_client_files(*paths)
        assert await pair.server_paths_exists(*paths) is False
        await pair.client.upload("foo")
        assert await pair.server_paths_exists(*paths)


@pytest.mark.asyncio
async def test_upload_folder_into(pair_factory):
    async with pair_factory() as pair:
        paths = ["foo/bar", "foo/baz"]
        await pair.make_client_files(*paths)
        assert await pair.server_paths_exists("bar", "baz") is False
        await pair.client.upload("foo", write_into=True)
        assert await pair.server_paths_exists("bar", "baz")


@pytest.mark.asyncio
async def test_upload_folder_into_another(pair_factory):
    async with pair_factory() as pair:
        paths = ["foo/bar", "foo/baz"]
        await pair.make_client_files(*paths)
        assert await pair.server_paths_exists("bar/bar", "bar/baz") is False
        await pair.client.upload("foo", "bar", write_into=True)
        assert await pair.server_paths_exists("bar/bar", "bar/baz")


@pytest.mark.asyncio
async def test_download_folder(pair_factory):
    async with pair_factory() as pair:
        paths = ["foo/bar", "foo/baz"]
        await pair.make_server_files(*paths)
        assert await pair.client_paths_exists(*paths) is False
        await pair.client.download("foo")
        assert await pair.client_paths_exists(*paths)


@pytest.mark.asyncio
async def test_download_folder_into(pair_factory):
    async with pair_factory() as pair:
        paths = ["foo/bar", "foo/baz"]
        await pair.make_server_files(*paths)
        assert await pair.client_paths_exists("bar", "baz") is False
        await pair.client.download("foo", write_into=True)
        assert await pair.client_paths_exists("bar", "baz")


@pytest.mark.asyncio
async def test_download_folder_into_another(pair_factory):
    async with pair_factory() as pair:
        paths = ["foo/bar", "foo/baz"]
        await pair.make_server_files(*paths)
        assert await pair.client_paths_exists("bar/bar", "bar/baz") is False
        await pair.client.download("foo", "bar", write_into=True)
        assert await pair.client_paths_exists("bar/bar", "bar/baz")


@pytest.mark.asyncio
async def test_upload_file_over(pair_factory):
    async with pair_factory() as pair:
        await pair.make_client_files("foo", size=1, atom=b"client")
        await pair.make_server_files("foo", size=1, atom=b"server")
        async with pair.client.download_stream("foo") as stream:
            assert await stream.read() == b"server"
        await pair.client.upload("foo")
        async with pair.client.download_stream("foo") as stream:
            assert await stream.read() == b"client"


@pytest.mark.asyncio
async def test_download_file_over(pair_factory):
    async with pair_factory() as pair:
        await pair.make_client_files("foo", size=1, atom=b"client")
        await pair.make_server_files("foo", size=1, atom=b"server")
        async with pair.client.path_io.open(PurePosixPath("foo")) as f:
            assert await f.read() == b"client"
        await pair.client.download("foo")
        async with pair.client.path_io.open(PurePosixPath("foo")) as f:
            assert await f.read() == b"server"


@pytest.mark.asyncio
async def test_upload_file_write_into(pair_factory):
    async with pair_factory() as pair:
        await pair.make_client_files("foo", size=1, atom=b"client")
        await pair.make_server_files("bar", size=1, atom=b"server")
        async with pair.client.download_stream("bar") as stream:
            assert await stream.read() == b"server"
        await pair.client.upload("foo", "bar", write_into=True)
        async with pair.client.download_stream("bar") as stream:
            assert await stream.read() == b"client"


@pytest.mark.asyncio
async def test_upload_tree(pair_factory):
    async with pair_factory() as pair:
        await pair.make_client_files("foo/bar/baz", size=1, atom=b"client")
        await pair.client.upload("foo", "bar", write_into=True)
        files = await pair.client.list(recursive=True)
        assert len(files) == 3


@pytest.mark.asyncio
async def test_download_file_write_into(pair_factory):
    async with pair_factory() as pair:
        await pair.make_client_files("foo", size=1, atom=b"client")
        await pair.make_server_files("bar", size=1, atom=b"server")
        async with pair.client.path_io.open(PurePosixPath("foo")) as f:
            assert await f.read() == b"client"
        await pair.client.download("bar", "foo", write_into=True)
        async with pair.client.path_io.open(PurePosixPath("foo")) as f:
            assert await f.read() == b"server"


@pytest.mark.asyncio
async def test_upload_file_os_error(pair_factory, Server,
                                    expect_codes_in_exception):

    class OsErrorPathIO(aioftp.MemoryPathIO):

        @aioftp.pathio.universal_exception
        async def write(self, fout, data):
            raise OSError("test os error")

    s = Server(path_io_factory=OsErrorPathIO)
    async with pair_factory(None, s) as pair:
        with expect_codes_in_exception("451"):
            async with pair.client.upload_stream("foo") as stream:
                await stream.write(b"foobar")


@pytest.mark.asyncio
async def test_upload_path_unreachable(pair_factory,
                                       expect_codes_in_exception):
    async with pair_factory() as pair:
        with expect_codes_in_exception("550"):
            async with pair.client.upload_stream("foo/bar/foo") as stream:
                await stream.write(b"foobar")


@pytest.mark.asyncio
async def test_stat_when_no_mlst(pair_factory):
    class CustomServer(aioftp.Server):
        def __getattribute__(self, name):
            if name == "mlst":
                raise AttributeError
            return super().__getattribute__(name)

    async with pair_factory(server_factory=CustomServer) as pair:
        await pair.make_server_files("foo")
        info = await pair.client.stat("foo")
        assert info["type"] == "file"



import ssl
import collections
import contextlib
import tempfile
import asyncio
import math
import time
import functools
import socket
from pathlib import Path

import pytest
import trustme
from async_timeout import timeout

import aioftp
from siosocks.io.asyncio import socks_server_handler


# No ssl tests since https://bugs.python.org/issue36098
ca = trustme.CA()
server_cert = ca.issue_server_cert("127.0.0.1", "::1")

ssl_server = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
server_cert.configure_cert(ssl_server)

ssl_client = ssl.create_default_context(ssl.Purpose.SERVER_AUTH)
ca.configure_trust(ssl_client)


class Container:
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs


@pytest.fixture
def Client():
    return Container


@pytest.fixture
def Server():
    return Container


def _wrap_with_defaults(kwargs):
    test_defaults = dict(
        path_io_factory=aioftp.MemoryPathIO,
    )
    return collections.ChainMap(kwargs, test_defaults)


@pytest.fixture(params=["127.0.0.1", "::1"])
def pair_factory(request):

    class Factory:

        def __init__(self, client=None, server=None, *,
                     connected=True, logged=True, do_quit=True,
                     host=request.param,
                     server_factory=aioftp.Server,
                     client_factory=aioftp.Client):
            if client is None:
                client = Container()
            self.client = client_factory(*client.args,
                                         **_wrap_with_defaults(client.kwargs))
            if server is None:
                server = Container()
            self.server = server_factory(*server.args,
                                         **_wrap_with_defaults(server.kwargs))
            self.connected = connected
            self.logged = logged
            self.do_quit = do_quit
            self.host = host
            self.timeout = timeout(1)

        async def make_server_files(self, *paths, size=None, atom=b"-"):
            if size is None:
                size = aioftp.DEFAULT_BLOCK_SIZE * 3
            data = atom * size
            for p in paths:
                await self.client.make_directory(Path(p).parent)
                async with self.client.upload_stream(p) as stream:
                    await stream.write(data)

        async def make_client_files(self, *paths, size=None, atom=b"-"):
            if size is None:
                size = aioftp.DEFAULT_BLOCK_SIZE * 3
            data = atom * size
            for p in map(Path, paths):
                await self.client.path_io.mkdir(p.parent, parents=True,
                                                exist_ok=True)
                async with self.client.path_io.open(p, mode="wb") as f:
                    await f.write(data)

        async def server_paths_exists(self, *paths):
            values = []
            for p in paths:
                values.append(await self.client.exists(p))
            if all(values):
                return True
            if any(values):
                raise ValueError("Mixed exists/not exists list")
            return False

        async def client_paths_exists(self, *paths):
            values = []
            for p in paths:
                values.append(await self.client.path_io.exists(Path(p)))
            if all(values):
                return True
            if any(values):
                raise ValueError("Mixed exists/not exists list")
            return False

        async def __aenter__(self):
            self.timeout.__enter__()
            await self.server.start(host=self.host)
            if self.connected:
                await self.client.connect(self.server.server_host,
                                          self.server.server_port)
                if self.logged:
                    await self.client.login()
            return self

        async def __aexit__(self, *exc_info):
            if self.connected and self.do_quit:
                await self.client.quit()
            self.client.close()
            await self.server.close()
            self.timeout.__exit__(*exc_info)

    return Factory


@pytest.fixture
def expect_codes_in_exception():
    @contextlib.contextmanager
    def context(*codes):
        try:
            yield
        except aioftp.StatusCodeError as e:
            assert set(e.received_codes) == set(codes)
        else:
            raise RuntimeError("There was no exception")
    return context


@pytest.fixture(params=[aioftp.MemoryPathIO, aioftp.PathIO,
                        aioftp.AsyncPathIO])
def path_io(request):
    return request.param()


@pytest.fixture
def temp_dir(path_io):
    if isinstance(path_io, aioftp.MemoryPathIO):
        yield Path("/")
    else:
        with tempfile.TemporaryDirectory() as name:
            yield Path(name)


class Sleep:

    def __init__(self):
        self.delay = 0
        self.first_sleep = None

    async def sleep(self, delay, result=None, **kwargs):
        if self.first_sleep is None:
            self.first_sleep = time.monotonic()
        delay = (time.monotonic() - self.first_sleep) + delay
        self.delay = max(self.delay, delay)
        return result

    def is_close(self, delay, *, rel_tol=0.05, abs_tol=0.5):
        ok = math.isclose(self.delay, delay, rel_tol=rel_tol, abs_tol=abs_tol)
        if not ok:
            print(f"latest sleep: {self.delay}; expected delay: "
                  f"{delay}; rel: {rel_tol}")
        return ok


@pytest.fixture
def skip_sleep(monkeypatch):
    with monkeypatch.context() as m:
        sleeper = Sleep()
        m.setattr(asyncio, "sleep", sleeper.sleep)
        yield sleeper


@pytest.fixture(params=[("127.0.0.1", socket.AF_INET),
                        ("::1", socket.AF_INET6)])
async def socks(request, unused_tcp_port):
    handler = functools.partial(
        socks_server_handler,
        allowed_versions={5},
        username="foo",
        password="bar",
    )
    Socks = collections.namedtuple("Socks", "host port server")
    host, family = request.param
    port = unused_tcp_port
    server = await asyncio.start_server(handler, host=host, port=port,
                                        family=family)
    yield Socks(host, port, server)
    server.close()
    await server.wait_closed()



import asyncio
import functools
import collections
import operator
import io
import time
import stat
import sys
import pathlib
import abc

from .common import (with_timeout, AsyncStreamIterator, DEFAULT_BLOCK_SIZE,
                     AbstractAsyncLister)
from . import errors


__all__ = (
    "AbstractPathIO",
    "PathIO",
    "AsyncPathIO",
    "MemoryPathIO",
    "PathIONursery",
)


class AsyncPathIOContext:
    """
    Async pathio context.

    Usage:
    ::

        >>> async with pathio.open(filename) as file_in:
        ...     async for block in file_in.iter_by_block(size):
        ...         # do

    or borring:
    ::

        >>> file = await pathio.open(filename)
        ... data = await file.read(size)
        ... await file.write(data)
        ... await file.close()

    """
    def __init__(self, pathio, args, kwargs):
        self.close = None
        self.pathio = pathio
        self.args = args
        self.kwargs = kwargs

    async def __aenter__(self):
        self.file = await self.pathio._open(*self.args, **self.kwargs)
        self.seek = functools.partial(self.pathio.seek, self.file)
        self.write = functools.partial(self.pathio.write, self.file)
        self.read = functools.partial(self.pathio.read, self.file)
        self.close = functools.partial(self.pathio.close, self.file)
        return self

    async def __aexit__(self, *args):
        if self.close is not None:
            await self.close()

    def __await__(self):
        return self.__aenter__().__await__()

    def iter_by_block(self, count=DEFAULT_BLOCK_SIZE):
        return AsyncStreamIterator(lambda: self.read(count))


def universal_exception(coro):
    """
    Decorator. Reraising any exception (except `CancelledError` and
    `NotImplementedError`) with universal exception
    :py:class:`aioftp.PathIOError`
    """
    @functools.wraps(coro)
    async def wrapper(*args, **kwargs):
        try:
            return await coro(*args, **kwargs)
        except (asyncio.CancelledError, NotImplementedError,
                StopAsyncIteration):
            raise
        except Exception:
            raise errors.PathIOError(reason=sys.exc_info())

    return wrapper


class PathIONursery:

    def __init__(self, factory):
        self.factory = factory
        self.state = None

    def __call__(self, *args, **kwargs):
        instance = self.factory(*args, state=self.state, **kwargs)
        if self.state is None:
            self.state = instance.state
        return instance


def defend_file_methods(coro):
    """
    Decorator. Raises exception when file methods called with wrapped by
    :py:class:`aioftp.AsyncPathIOContext` file object.
    """
    @functools.wraps(coro)
    async def wrapper(self, file, *args, **kwargs):
        if isinstance(file, AsyncPathIOContext):
            raise ValueError("Native path io file methods can not be used "
                             "with wrapped file object")
        return await coro(self, file, *args, **kwargs)
    return wrapper


class AbstractPathIO(abc.ABC):
    """
    Abstract class for path io operations.

    :param timeout: timeout used by `with_timeout` decorator
    :type timeout: :py:class:`float`, :py:class:`int` or `None`

    :param connection: server connection that is accessing this PathIO
    :type connection: :py:class:`aioftp.Connection`

    :param state: shared pathio state per server
    """
    def __init__(self, timeout=None, connection=None, state=None):
        self.timeout = timeout
        self.connection = connection

    @property
    def state(self):
        """
        Shared pathio state per server
        """

    @universal_exception
    @abc.abstractmethod
    async def exists(self, path):
        """
        :py:func:`asyncio.coroutine`

        Check if path exists

        :param path: path to check
        :type path: :py:class:`pathlib.Path`

        :rtype: :py:class:`bool`
        """

    @universal_exception
    @abc.abstractmethod
    async def is_dir(self, path):
        """
        :py:func:`asyncio.coroutine`

        Check if path is directory

        :param path: path to check
        :type path: :py:class:`pathlib.Path`

        :rtype: :py:class:`bool`
        """

    @universal_exception
    @abc.abstractmethod
    async def is_file(self, path):
        """
        :py:func:`asyncio.coroutine`

        Check if path is file

        :param path: path to check
        :type path: :py:class:`pathlib.Path`

        :rtype: :py:class:`bool`
        """

    @universal_exception
    @abc.abstractmethod
    async def mkdir(self, path, *, parents=False, exist_ok=False):
        """
        :py:func:`asyncio.coroutine`

        Make directory

        :param path: path to create
        :type path: :py:class:`pathlib.Path`

        :param parents: create parents is does not exists
        :type parents: :py:class:`bool`

        :param exist_ok: do not raise exception if directory already exists
        :type exist_ok: :py:class:`bool`
        """

    @universal_exception
    @abc.abstractmethod
    async def rmdir(self, path):
        """
        :py:func:`asyncio.coroutine`

        Remove directory

        :param path: path to remove
        :type path: :py:class:`pathlib.Path`
        """

    @universal_exception
    @abc.abstractmethod
    async def unlink(self, path):
        """
        :py:func:`asyncio.coroutine`

        Remove file

        :param path: path to remove
        :type path: :py:class:`pathlib.Path`
        """

    @abc.abstractmethod
    def list(self, path):
        """
        Create instance of subclass of :py:class:`aioftp.AbstractAsyncLister`.
        You should subclass and implement `__anext__` method
        for :py:class:`aioftp.AbstractAsyncLister` and return new instance.

        :param path: path to list
        :type path: :py:class:`pathlib.Path`

        :rtype: :py:class:`aioftp.AbstractAsyncLister`

        Usage:
        ::

            >>> async for p in pathio.list(path):
            ...     # do

        or borring instance of :py:class:`list`:
        ::

            >>> paths = await pathio.list(path)
            >>> paths
            [path, path, path, ...]

        """

    @universal_exception
    @abc.abstractmethod
    async def stat(self, path):
        """
        :py:func:`asyncio.coroutine`

        Get path stats

        :param path: path, which stats need
        :type path: :py:class:`pathlib.Path`

        :return: path stats. For proper work you need only this stats:
          st_size, st_mtime, st_ctime, st_nlink, st_mode
        :rtype: same as :py:class:`os.stat_result`
        """

    @universal_exception
    @abc.abstractmethod
    async def _open(self, path, mode):
        """
        :py:func:`asyncio.coroutine`

        Open file. You should implement "mode" argument, which can be:
        "rb", "wb", "ab" (read, write, append. all binary). Return type depends
        on implementation, anyway the only place you need this file-object
        is in your implementation of read, write and close

        :param path: path to create
        :type path: :py:class:`pathlib.Path`

        :param mode: specifies the mode in which the file is opened ("rb",
            "wb", "ab", "r+b" (read, write, append, read/write, all binary))
        :type mode: :py:class:`str`

        :return: file-object
        """

    def open(self, *args, **kwargs):
        """
        Create instance of :py:class:`aioftp.pathio.AsyncPathIOContext`,
        parameters passed to :py:meth:`aioftp.AbstractPathIO._open`

        :rtype: :py:class:`aioftp.pathio.AsyncPathIOContext`
        """
        return AsyncPathIOContext(self, args, kwargs)

    @universal_exception
    @defend_file_methods
    @abc.abstractmethod
    async def seek(self, file, offset, whence=io.SEEK_SET):
        """
        :py:func:`asyncio.coroutine`

        Change the stream position to the given byte `offset`. Same behaviour
        as :py:meth:`io.IOBase.seek`

        :param file: file-object from :py:class:`aioftp.AbstractPathIO.open`

        :param offset: relative byte offset
        :type offset: :py:class:`int`

        :param whence: base position for offset
        :type whence: :py:class:`int`
        """

    @universal_exception
    @defend_file_methods
    @abc.abstractmethod
    async def write(self, file, data):
        """
        :py:func:`asyncio.coroutine`

        Write some data to file

        :param file: file-object from :py:class:`aioftp.AbstractPathIO.open`

        :param data: data to write
        :type data: :py:class:`bytes`
        """

    @universal_exception
    @defend_file_methods
    @abc.abstractmethod
    async def read(self, file, block_size):
        """
        :py:func:`asyncio.coroutine`

        Read some data from file

        :param file: file-object from :py:class:`aioftp.AbstractPathIO.open`

        :param block_size: bytes count to read
        :type block_size: :py:class:`int`

        :rtype: :py:class:`bytes`
        """

    @universal_exception
    @defend_file_methods
    @abc.abstractmethod
    async def close(self, file):
        """
        :py:func:`asyncio.coroutine`

        Close file

        :param file: file-object from :py:class:`aioftp.AbstractPathIO.open`
        """

    @universal_exception
    @abc.abstractmethod
    async def rename(self, source, destination):
        """
        :py:func:`asyncio.coroutine`

        Rename path

        :param source: rename from
        :type source: :py:class:`pathlib.Path`

        :param destination: rename to
        :type destination: :py:class:`pathlib.Path`
        """


class PathIO(AbstractPathIO):
    """
    Blocking path io. Directly based on :py:class:`pathlib.Path` methods.
    """

    @universal_exception
    async def exists(self, path):
        return path.exists()

    @universal_exception
    async def is_dir(self, path):
        return path.is_dir()

    @universal_exception
    async def is_file(self, path):
        return path.is_file()

    @universal_exception
    async def mkdir(self, path, *, parents=False, exist_ok=False):
        return path.mkdir(parents=parents, exist_ok=exist_ok)

    @universal_exception
    async def rmdir(self, path):
        return path.rmdir()

    @universal_exception
    async def unlink(self, path):
        return path.unlink()

    def list(self, path):

        class Lister(AbstractAsyncLister):
            iter = None

            @universal_exception
            async def __anext__(self):
                if self.iter is None:
                    self.iter = path.glob("*")
                try:
                    return next(self.iter)
                except StopIteration:
                    raise StopAsyncIteration

        return Lister(timeout=self.timeout)

    @universal_exception
    async def stat(self, path):
        return path.stat()

    @universal_exception
    async def _open(self, path, *args, **kwargs):
        return path.open(*args, **kwargs)

    @universal_exception
    @defend_file_methods
    async def seek(self, file, *args, **kwargs):
        return file.seek(*args, **kwargs)

    @universal_exception
    @defend_file_methods
    async def write(self, file, *args, **kwargs):
        return file.write(*args, **kwargs)

    @universal_exception
    @defend_file_methods
    async def read(self, file, *args, **kwargs):
        return file.read(*args, **kwargs)

    @universal_exception
    @defend_file_methods
    async def close(self, file):
        return file.close()

    @universal_exception
    async def rename(self, source, destination):
        return source.rename(destination)


def _blocking_io(f):
    @functools.wraps(f)
    async def wrapper(self, *args, **kwargs):
        return await asyncio.get_event_loop().run_in_executor(
            self.executor,
            functools.partial(f, self, *args, **kwargs),
        )
    return wrapper


class AsyncPathIO(AbstractPathIO):
    """
    Non-blocking path io. Based on
    :py:meth:`asyncio.BaseEventLoop.run_in_executor` and
    :py:class:`pathlib.Path` methods. It's really slow, so it's better to avoid
    usage of this path io layer.

    :param executor: executor for running blocking tasks
    :type executor: :py:class:`concurrent.futures.Executor`
    """
    def __init__(self, *args, executor=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.executor = executor

    @universal_exception
    @with_timeout
    @_blocking_io
    def exists(self, path):
        return path.exists()

    @universal_exception
    @with_timeout
    @_blocking_io
    def is_dir(self, path):
        return path.is_dir()

    @universal_exception
    @with_timeout
    @_blocking_io
    def is_file(self, path):
        return path.is_file()

    @universal_exception
    @with_timeout
    @_blocking_io
    def mkdir(self, path, *, parents=False, exist_ok=False):
        return path.mkdir(parents=parents, exist_ok=exist_ok)

    @universal_exception
    @with_timeout
    @_blocking_io
    def rmdir(self, path):
        return path.rmdir()

    @universal_exception
    @with_timeout
    @_blocking_io
    def unlink(self, path):
        return path.unlink()

    def list(self, path):

        class Lister(AbstractAsyncLister):
            iter = None

            def __init__(self, *args, executor=None, **kwargs):
                super().__init__(*args, **kwargs)
                self.executor = executor

            def worker(self):
                try:
                    return next(self.iter)
                except StopIteration:
                    raise StopAsyncIteration

            @universal_exception
            @with_timeout
            @_blocking_io
            def __anext__(self):
                if self.iter is None:
                    self.iter = path.glob("*")
                return self.worker()

        return Lister(timeout=self.timeout, executor=self.executor)

    @universal_exception
    @with_timeout
    @_blocking_io
    def stat(self, path):
        return path.stat()

    @universal_exception
    @with_timeout
    @_blocking_io
    def _open(self, path, *args, **kwargs):
        return path.open(*args, **kwargs)

    @universal_exception
    @defend_file_methods
    @with_timeout
    @_blocking_io
    def seek(self, file, *args, **kwargs):
        return file.seek(*args, **kwargs)

    @universal_exception
    @defend_file_methods
    @with_timeout
    @_blocking_io
    def write(self, file, *args, **kwargs):
        return file.write(*args, **kwargs)

    @universal_exception
    @defend_file_methods
    @with_timeout
    @_blocking_io
    def read(self, file, *args, **kwargs):
        return file.read(*args, **kwargs)

    @universal_exception
    @defend_file_methods
    @with_timeout
    @_blocking_io
    def close(self, file):
        return file.close()

    @universal_exception
    @with_timeout
    @_blocking_io
    def rename(self, source, destination):
        return source.rename(destination)


class Node:

    def __init__(self, type, name, ctime=None, mtime=None, *, content):
        self.type = type
        self.name = name
        self.ctime = ctime or int(time.time())
        self.mtime = mtime or int(time.time())
        self.content = content

    def __repr__(self):
        return f"{self.__class__.__name__}(type={self.type!r}, " \
               f"name={self.name!r}, ctime={self.ctime!r}, " \
               f"mtime={self.mtime!r}, content={self.content!r})"


class MemoryPathIO(AbstractPathIO):
    """
    Non-blocking path io. Based on in-memory tree. It is just proof of concept
    and probably not so fast as it can be.
    """

    Stats = collections.namedtuple(
        "Stats",
        (
            "st_size",
            "st_ctime",
            "st_mtime",
            "st_nlink",
            "st_mode",
        )
    )

    def __init__(self, *args, state=None, cwd=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.cwd = pathlib.PurePosixPath(cwd or "/")
        if state is None:
            self.fs = [Node("dir", "/", content=[])]
        else:
            self.fs = state

    @property
    def state(self):
        return self.fs

    def __repr__(self):
        return repr(self.fs)

    def _absolute(self, path):
        if not path.is_absolute():
            path = self.cwd / path
        return path

    def get_node(self, path):
        nodes = self.fs
        node = None
        path = self._absolute(path)
        for part in path.parts:
            if not isinstance(nodes, list):
                return
            for node in nodes:
                if node.name == part:
                    nodes = node.content
                    break
            else:
                return
        return node

    @universal_exception
    async def exists(self, path):
        return self.get_node(path) is not None

    @universal_exception
    async def is_dir(self, path):
        node = self.get_node(path)
        return not (node is None or node.type != "dir")

    @universal_exception
    async def is_file(self, path):
        node = self.get_node(path)
        return not (node is None or node.type != "file")

    @universal_exception
    async def mkdir(self, path, *, parents=False, exist_ok=False):
        path = self._absolute(path)
        node = self.get_node(path)
        if node:
            if node.type != "dir" or not exist_ok:
                raise FileExistsError
        elif not parents:
            parent = self.get_node(path.parent)
            if parent is None:
                raise FileNotFoundError
            elif parent.type != "dir":
                raise NotADirectoryError
            node = Node("dir", path.name, content=[])
            parent.content.append(node)
        else:
            nodes = self.fs
            for part in path.parts:
                if isinstance(nodes, list):
                    for node in nodes:
                        if node.name == part:
                            nodes = node.content
                            break
                    else:
                        node = Node("dir", part, content=[])
                        nodes.append(node)
                        nodes = node.content
                else:
                    raise NotADirectoryError

    @universal_exception
    async def rmdir(self, path):
        node = self.get_node(path)
        if node is None:
            raise FileNotFoundError
        elif node.type != "dir":
            raise NotADirectoryError
        elif node.content:
            raise OSError("Directory not empty")
        else:
            parent = self.get_node(path.parent)
            for i, node in enumerate(parent.content):
                if node.name == path.name:
                    break
            parent.content.pop(i)

    @universal_exception
    async def unlink(self, path):
        node = self.get_node(path)
        if node is None:
            raise FileNotFoundError
        elif node.type != "file":
            raise IsADirectoryError
        else:
            parent = self.get_node(path.parent)
            for i, node in enumerate(parent.content):
                if node.name == path.name:
                    break
            parent.content.pop(i)

    def list(self, path):

        class Lister(AbstractAsyncLister):
            iter = None

            @universal_exception
            async def __anext__(cls):
                if cls.iter is None:
                    node = self.get_node(path)
                    if node is None or node.type != "dir":
                        cls.iter = iter(())
                    else:
                        names = map(operator.attrgetter("name"), node.content)
                        paths = map(lambda name: path / name, names)
                        cls.iter = iter(paths)
                try:
                    return next(cls.iter)
                except StopIteration:
                    raise StopAsyncIteration

        return Lister(timeout=self.timeout)

    @universal_exception
    async def stat(self, path):
        node = self.get_node(path)
        if node is None:
            raise FileNotFoundError
        else:
            if node.type == "file":
                size = len(node.content.getbuffer())
                mode = stat.S_IFREG | 0o666
            else:
                size = 0
                mode = stat.S_IFDIR | 0o777
            return MemoryPathIO.Stats(
                size,
                node.ctime,
                node.mtime,
                1,
                mode,
            )

    @universal_exception
    async def _open(self, path, mode="rb", *args, **kwargs):
        if mode == "rb":
            node = self.get_node(path)
            if node is None:
                raise FileNotFoundError
            file_like = node.content
            file_like.seek(0, io.SEEK_SET)
        elif mode in ("wb", "ab", "r+b"):
            node = self.get_node(path)
            if node is None:
                parent = self.get_node(path.parent)
                if parent is None or parent.type != "dir":
                    raise FileNotFoundError
                new_node = Node("file", path.name, content=io.BytesIO())
                parent.content.append(new_node)
                file_like = new_node.content
            elif node.type != "file":
                raise IsADirectoryError
            else:
                if mode == "wb":
                    file_like = node.content = io.BytesIO()
                elif mode == "ab":
                    file_like = node.content
                    file_like.seek(0, io.SEEK_END)
                elif mode == "r+b":
                    file_like = node.content
                    file_like.seek(0, io.SEEK_SET)
        else:
            raise ValueError(f"invalid mode: {mode}")
        return file_like

    @universal_exception
    @defend_file_methods
    async def seek(self, file, *args, **kwargs):
        return file.seek(*args, **kwargs)

    @universal_exception
    @defend_file_methods
    async def write(self, file, *args, **kwargs):
        file.write(*args, **kwargs)
        file.mtime = int(time.time())

    @universal_exception
    @defend_file_methods
    async def read(self, file, *args, **kwargs):
        return file.read(*args, **kwargs)

    @universal_exception
    @defend_file_methods
    async def close(self, file):
        pass

    @universal_exception
    async def rename(self, source, destination):
        if source != destination:
            sparent = self.get_node(source.parent)
            dparent = self.get_node(destination.parent)
            snode = self.get_node(source)
            if None in (snode, dparent):
                raise FileNotFoundError
            for i, node in enumerate(sparent.content):
                if node.name == source.name:
                    sparent.content.pop(i)
            snode.name = destination.name
            for i, node in enumerate(dparent.content):
                if node.name == destination.name:
                    dparent.content[i] = snode
                    break
            else:
                dparent.content.append(snode)



import asyncio
from functools import reduce
from pathlib import Path

import pytest

import aioftp


@pytest.mark.asyncio
async def test_patched_sleep(skip_sleep):
    await asyncio.sleep(10)
    assert skip_sleep.is_close(10)


SIZE = 3 * 100 * 1024  # 300KiB


@pytest.mark.parametrize("times", [10, 20, 30])
@pytest.mark.parametrize("type", ["read", "write"])
@pytest.mark.parametrize("direction", ["download", "upload"])
@pytest.mark.asyncio
async def test_client_side_throttle(pair_factory, skip_sleep, times, type,
                                    direction):
    async with pair_factory() as pair:
        await pair.make_server_files("foo", size=SIZE)
        await pair.make_client_files("foo", size=SIZE)
        getattr(pair.client.throttle, type).limit = SIZE / times
        await getattr(pair.client, direction)("foo")
        if (type, direction) in {("read", "download"), ("write", "upload")}:
            assert skip_sleep.is_close(times)
        else:
            assert skip_sleep.is_close(0)


@pytest.mark.parametrize("times", [10, 20, 30])
@pytest.mark.parametrize("users", [1, 2, 3])
@pytest.mark.parametrize("throttle_direction", ["read", "write"])
@pytest.mark.parametrize("data_direction", ["download", "upload"])
@pytest.mark.parametrize("throttle_level", ["throttle",
                                            "throttle_per_connection"])
@pytest.mark.asyncio
async def test_server_side_throttle(pair_factory, skip_sleep, times, users,
                                    throttle_direction, data_direction,
                                    throttle_level):
    async with pair_factory() as pair:
        names = []
        for i in range(users):
            name = f"foo{i}"
            names.append(name)
            await pair.make_server_files(name, size=SIZE)
        throttle = reduce(getattr, [throttle_level, throttle_direction],
                          pair.server)
        throttle.limit = SIZE / times
        clients = []
        for name in names:
            c = aioftp.Client(path_io_factory=aioftp.MemoryPathIO)
            async with c.path_io.open(Path(name), "wb") as f:
                await f.write(b"-" * SIZE)
            await c.connect(pair.server.server_host, pair.server.server_port)
            await c.login()
            clients.append(c)
        coros = [getattr(c, data_direction)(n) for c, n in zip(clients, names)]
        await asyncio.gather(*coros)
        await asyncio.gather(*[c.quit() for c in clients])
    throttled = {("read", "upload"), ("write", "download")}
    if (throttle_direction, data_direction) not in throttled:
        assert skip_sleep.is_close(0)
    else:
        t = times
        if throttle_level == "throttle":  # global
            t *= users
        assert skip_sleep.is_close(t)



import pytest

import aioftp


@pytest.mark.asyncio
async def test_client_list_override(pair_factory, expect_codes_in_exception):
    async with pair_factory(logged=False, do_quit=False) as pair:
        with expect_codes_in_exception("503"):
            await pair.client.get_current_directory()


@pytest.mark.asyncio
async def test_anonymous_login(pair_factory):
    async with pair_factory():
        pass


@pytest.mark.asyncio
async def test_login_with_login_data(pair_factory):
    async with pair_factory(logged=False) as pair:
        await pair.client.login("foo", "bar")


@pytest.mark.asyncio
async def test_login_with_login_and_no_password(pair_factory, Server):
    s = Server([aioftp.User("foo")])
    async with pair_factory(None, s, logged=False) as pair:
        await pair.client.login("foo")


@pytest.mark.asyncio
async def test_login_with_login_and_password(pair_factory, Server):
    s = Server([aioftp.User("foo", "bar")])
    async with pair_factory(None, s, logged=False) as pair:
        await pair.client.login("foo", "bar")


@pytest.mark.asyncio
async def test_login_with_login_and_password_no_such_user(
        pair_factory, Server, expect_codes_in_exception):
    s = Server([aioftp.User("foo", "bar")])
    async with pair_factory(None, s, logged=False) as pair:
        with expect_codes_in_exception("530"):
            await pair.client.login("fo", "bar")


@pytest.mark.asyncio
async def test_login_with_login_and_password_bad_password(
        pair_factory, Server, expect_codes_in_exception):
    s = Server([aioftp.User("foo", "bar")])
    async with pair_factory(None, s, logged=False) as pair:
        with expect_codes_in_exception("530"):
            await pair.client.login("foo", "baz")


@pytest.mark.asyncio
async def test_pass_after_login(pair_factory, Server,
                                expect_codes_in_exception):
    s = Server([aioftp.User("foo", "bar")])
    async with pair_factory(None, s, logged=False) as pair:
        await pair.client.login("foo", "bar")
        with expect_codes_in_exception("503"):
            await pair.client.command("PASS baz", ("230", "33x"))



import pytest


@pytest.mark.parametrize("offset", [0, 3, 10])
@pytest.mark.asyncio
async def test_restart_retr(pair_factory, offset):
    async with pair_factory() as pair:
        atom = b"foobar"
        name = "foo.txt"
        await pair.make_server_files(name, size=1, atom=atom)
        async with pair.client.download_stream(name, offset=offset) as stream:
            assert await stream.read() == atom[offset:]


@pytest.mark.parametrize("offset", [1, 3, 10])
@pytest.mark.parametrize("method", ["upload_stream", "append_stream"])
@pytest.mark.asyncio
async def test_restart_stor_appe(pair_factory, offset, method):
    async with pair_factory() as pair:
        atom = b"foobar"
        name = "foo.txt"
        insert = b"123"
        expect = atom[:offset] + b"\x00" * (offset - len(atom)) + insert + \
            atom[offset + len(insert):]
        await pair.make_server_files(name, size=1, atom=atom)
        stream_factory = getattr(pair.client, method)
        async with stream_factory(name, offset=offset) as stream:
            await stream.write(b"123")
        async with pair.client.download_stream(name) as stream:
            assert await stream.read() == expect


@pytest.mark.asyncio
async def test_restart_reset(pair_factory):
    async with pair_factory() as pair:
        atom = b"foobar"
        name = "foo.txt"
        await pair.make_server_files(name, size=1, atom=atom)
        await pair.client.command("REST 3", "350")
        async with pair.client.download_stream("foo.txt") as stream:
            assert await stream.read() == atom


@pytest.mark.asyncio
async def test_restart_syntax_error(pair_factory, expect_codes_in_exception):
    async with pair_factory() as pair:
        with expect_codes_in_exception("501"):
            await pair.client.command("REST 3abc", "350")



import functools

import pytest

import aioftp


@pytest.mark.asyncio
async def test_multiply_connections_no_limits(pair_factory):
    Client = functools.partial(aioftp.Client,
                               path_io_factory=aioftp.MemoryPathIO)
    async with pair_factory() as pair:
        s = pair.server
        clients = [Client() for _ in range(4)]
        for c in clients:
            await c.connect(s.server_host, s.server_port)
            await c.login()
        for c in clients:
            await c.quit()


@pytest.mark.asyncio
async def test_multiply_connections_limited_error(pair_factory, Server,
                                                  expect_codes_in_exception):
    Client = functools.partial(aioftp.Client,
                               path_io_factory=aioftp.MemoryPathIO)
    s = Server(maximum_connections=4)
    async with pair_factory(None, s) as pair:
        s = pair.server
        clients = [Client() for _ in range(4)]
        for c in clients[:-1]:
            await c.connect(s.server_host, s.server_port)
            await c.login()
        with expect_codes_in_exception("421"):
            await clients[-1].connect(s.server_host, s.server_port)
        for c in clients[:-1]:
            await c.quit()


@pytest.mark.asyncio
async def test_multiply_user_commands(pair_factory, Server):
    s = Server(maximum_connections=1)
    async with pair_factory(None, s) as pair:
        for _ in range(10):
            await pair.client.login()


@pytest.mark.asyncio
async def test_multiply_connections_with_user_limited_error(
        pair_factory, Server, expect_codes_in_exception):
    Client = functools.partial(aioftp.Client,
                               path_io_factory=aioftp.MemoryPathIO)
    s = Server([aioftp.User("foo", maximum_connections=4)])
    async with pair_factory(None, s, connected=False) as pair:
        s = pair.server
        clients = [Client() for _ in range(5)]
        for c in clients[:-1]:
            await c.connect(s.server_host, s.server_port)
            await c.login("foo")
        await clients[-1].connect(s.server_host, s.server_port)
        with expect_codes_in_exception("530"):
            await clients[-1].login("foo")
        for c in clients[:-1]:
            await c.quit()


@pytest.mark.asyncio
async def test_multiply_connections_relogin_balanced(
        pair_factory, Server, expect_codes_in_exception):
    Client = functools.partial(aioftp.Client,
                               path_io_factory=aioftp.MemoryPathIO)
    s = Server(maximum_connections=4)
    async with pair_factory(None, s, connected=False) as pair:
        s = pair.server
        clients = [Client() for _ in range(5)]
        for c in clients[:-1]:
            await c.connect(s.server_host, s.server_port)
            await c.login()
        await clients[0].quit()
        await clients[-1].connect(s.server_host, s.server_port)
        await clients[-1].login()
        for c in clients[1:]:
            await c.quit()



import pathlib

import pytest

import aioftp


@pytest.mark.asyncio
async def test_current_directory_simple(pair_factory):
    async with pair_factory() as pair:
        cwd = await pair.client.get_current_directory()
        assert cwd == pathlib.PurePosixPath("/")


@pytest.mark.asyncio
async def test_current_directory_not_default(pair_factory, Server):
    s = Server([aioftp.User(home_path="/home")])
    async with pair_factory(None, s) as pair:
        cwd = await pair.client.get_current_directory()
        assert cwd == pathlib.PurePosixPath("/home")


@pytest.mark.asyncio
async def test_mlsd(pair_factory):
    async with pair_factory() as pair:
        await pair.make_server_files("test.txt")
        (path, stat), *_ = files = await pair.client.list()
        assert len(files) == 1
        assert path == pathlib.PurePosixPath("test.txt")
        assert stat["type"] == "file"


@pytest.mark.asyncio
async def test_resolving_double_dots(pair_factory):
    async with pair_factory() as pair:
        await pair.make_server_files("test.txt")

        async def f():
            cwd = await pair.client.get_current_directory()
            assert cwd == pathlib.PurePosixPath("/")
            (path, stat), *_ = files = await pair.client.list()
            assert len(files) == 1
            assert path == pathlib.PurePosixPath("test.txt")
            assert stat["type"] == "file"

        await f()
        await pair.client.change_directory("../../../")
        await f()
        await pair.client.change_directory("/a/../b/..")
        await f()



import asyncio
import itertools
import pathlib

import pytest

import aioftp


@pytest.mark.asyncio
async def test_abort_stor(pair_factory):
    async with pair_factory() as pair:
        stream = await pair.client.upload_stream("test.txt")
        with pytest.raises((ConnectionResetError, BrokenPipeError)):
            while True:
                await stream.write(b"-" * aioftp.DEFAULT_BLOCK_SIZE)
                await pair.client.abort()


class SlowReadMemoryPathIO(aioftp.MemoryPathIO):

    async def read(self, *args, **kwargs):
        await asyncio.sleep(0.01)
        return await super().read(*args, **kwargs)


@pytest.mark.asyncio
async def test_abort_retr(pair_factory, Server):
    s = Server(path_io_factory=SlowReadMemoryPathIO)
    async with pair_factory(None, s) as pair:
        await pair.make_server_files("test.txt")
        stream = await pair.client.download_stream("test.txt")
        for i in itertools.count():
            data = await stream.read(aioftp.DEFAULT_BLOCK_SIZE)
            if not data:
                stream.close()
                break
            if i == 0:
                await pair.client.abort()


@pytest.mark.asyncio
async def test_abort_retr_no_wait(pair_factory, Server,
                                  expect_codes_in_exception):
    s = Server(path_io_factory=SlowReadMemoryPathIO)
    async with pair_factory(None, s) as pair:
        await pair.make_server_files("test.txt")
        stream = await pair.client.download_stream("test.txt")
        with expect_codes_in_exception("426"):
            for i in itertools.count():
                data = await stream.read(aioftp.DEFAULT_BLOCK_SIZE)
                if not data:
                    await stream.finish()
                    break
                if i == 0:
                    await pair.client.abort(wait=False)


@pytest.mark.asyncio
async def test_nothing_to_abort(pair_factory):
    async with pair_factory() as pair:
        await pair.client.abort()


class SlowListMemoryPathIO(aioftp.MemoryPathIO):

    async def is_file(self, *a, **kw):
        return True

    def list(self, *args, **kwargs):

        class Lister(aioftp.AbstractAsyncLister):

            async def __anext__(cls):
                await asyncio.sleep(0.01)
                return pathlib.PurePath("/test.txt")

        return Lister()

    async def stat(self, *a, **kw):
        class Stat:
            st_size = 0
            st_mtime = 0
            st_ctime = 0
        return Stat


@pytest.mark.asyncio
async def test_mlsd_abort(pair_factory, Server):
    s = Server(path_io_factory=SlowListMemoryPathIO)
    async with pair_factory(None, s) as pair:
        cwd = await pair.client.get_current_directory()
        assert cwd == pathlib.PurePosixPath("/")
        async for path, info in pair.client.list():
            await pair.client.abort()
            break



import re
import collections
import pathlib
import logging
import datetime
from functools import partial

try:
    from siosocks.io.asyncio import open_connection
except ImportError:
    from asyncio import open_connection

from . import errors
from . import pathio
from .common import (
    ThrottleStreamIO,
    StreamThrottle,
    DEFAULT_PORT,
    wrap_with_container,
    END_OF_LINE,
    DEFAULT_USER,
    DEFAULT_PASSWORD,
    DEFAULT_ACCOUNT,
    DEFAULT_BLOCK_SIZE,
    AsyncListerMixin,
    async_enterable,
    setlocale,
    HALF_OF_YEAR_IN_SECONDS,
)

__all__ = (
    "BaseClient",
    "Client",
    "ClientSession",
    "DataConnectionThrottleStreamIO",
    "Code",
)
logger = logging.getLogger(__name__)


class Code(str):
    """
    Representation of server status code.
    """
    def matches(self, mask):
        """
        :param mask: Template for comparision. If mask symbol is not digit
            then it passes.
        :type mask: :py:class:`str`

        ::

            >>> Code("123").matches("1")
            True
            >>> Code("123").matches("1x3")
            True
        """
        return all(map(lambda m, c: not m.isdigit() or m == c, mask, self))


class DataConnectionThrottleStreamIO(ThrottleStreamIO):
    """
    Add `finish` method to :py:class:`aioftp.ThrottleStreamIO`, which is
    specific for data connection. This requires `client`.

    :param client: client class, which have :py:meth:`aioftp.Client.command`
    :type client: :py:class:`aioftp.BaseClient`

    :param *args: positional arguments passed to
        :py:class:`aioftp.ThrottleStreamIO`

    :param **kwargs: keyword arguments passed to
        :py:class:`aioftp.ThrottleStreamIO`
    """
    def __init__(self, client, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.client = client

    async def finish(self, expected_codes="2xx", wait_codes="1xx"):
        """
        :py:func:`asyncio.coroutine`

        Close connection and wait for `expected_codes` response from server
        passing `wait_codes`.

        :param expected_codes: tuple of expected codes or expected code
        :type expected_codes: :py:class:`tuple` of :py:class:`str` or
            :py:class:`str`

        :param wait_codes: tuple of wait codes or wait code
        :type wait_codes: :py:class:`tuple` of :py:class:`str` or
            :py:class:`str`
        """
        self.close()
        await self.client.command(None, expected_codes, wait_codes)

    async def __aexit__(self, exc_type, exc, tb):
        if exc is None:
            await self.finish()
        else:
            self.close()


class BaseClient:

    def __init__(self, *, socket_timeout=None,
                 read_speed_limit=None, write_speed_limit=None,
                 path_timeout=None, path_io_factory=pathio.PathIO,
                 encoding="utf-8", ssl=None, parse_list_line_custom=None,
                 **siosocks_asyncio_kwargs):
        self.socket_timeout = socket_timeout
        self.throttle = StreamThrottle.from_limits(
            read_speed_limit,
            write_speed_limit,
        )
        self.path_timeout = path_timeout
        self.path_io = path_io_factory(timeout=path_timeout)
        self.encoding = encoding
        self.stream = None
        self.ssl = ssl
        self.parse_list_line_custom = parse_list_line_custom
        self._open_connection = partial(open_connection, ssl=self.ssl,
                                        **siosocks_asyncio_kwargs)

    async def connect(self, host, port=DEFAULT_PORT):
        self.server_host = host
        self.server_port = port
        reader, writer = await self._open_connection(host, port)
        self.stream = ThrottleStreamIO(
            reader,
            writer,
            throttles={"_": self.throttle},
            timeout=self.socket_timeout,
        )

    def close(self):
        """
        Close connection.
        """
        if self.stream is not None:
            self.stream.close()

    async def parse_line(self):
        """
        :py:func:`asyncio.coroutine`

        Parsing server response line.

        :return: (code, line)
        :rtype: (:py:class:`aioftp.Code`, :py:class:`str`)

        :raises ConnectionResetError: if received data is empty (this
            means, that connection is closed)
        :raises asyncio.TimeoutError: if there where no data for `timeout`
            period
        """
        line = await self.stream.readline()
        if not line:
            self.stream.close()
            raise ConnectionResetError
        s = line.decode(encoding=self.encoding).rstrip()
        logger.info(s)
        return Code(s[:3]), s[3:]

    async def parse_response(self):
        """
        :py:func:`asyncio.coroutine`

        Parsing full server response (all lines).

        :return: (code, lines)
        :rtype: (:py:class:`aioftp.Code`, :py:class:`list` of :py:class:`str`)

        :raises aioftp.StatusCodeError: if received code does not matches all
            already received codes
        """
        code, rest = await self.parse_line()
        info = [rest]
        curr_code = code
        while rest.startswith("-") or not curr_code.isdigit():
            curr_code, rest = await self.parse_line()
            if curr_code.isdigit():
                info.append(rest)
                if curr_code != code:
                    raise errors.StatusCodeError(code, curr_code, info)
            else:
                info.append(curr_code + rest)
        return code, info

    def check_codes(self, expected_codes, received_code, info):
        """
        Checks if any of expected matches received.

        :param expected_codes: tuple of expected codes
        :type expected_codes: :py:class:`tuple`

        :param received_code: received code for matching
        :type received_code: :py:class:`aioftp.Code`

        :param info: list of response lines from server
        :type info: :py:class:`list`

        :raises aioftp.StatusCodeError: if received code does not matches any
            expected code
        """
        if not any(map(received_code.matches, expected_codes)):
            raise errors.StatusCodeError(expected_codes, received_code, info)

    async def command(self, command=None, expected_codes=(), wait_codes=()):
        """
        :py:func:`asyncio.coroutine`

        Basic command logic.

        1. Send command if not omitted.
        2. Yield response until no wait code matches.
        3. Check code for expected.

        :param command: command line
        :type command: :py:class:`str`

        :param expected_codes: tuple of expected codes or expected code
        :type expected_codes: :py:class:`tuple` of :py:class:`str` or
            :py:class:`str`

        :param wait_codes: tuple of wait codes or wait code
        :type wait_codes: :py:class:`tuple` of :py:class:`str` or
            :py:class:`str`
        """
        expected_codes = wrap_with_container(expected_codes)
        wait_codes = wrap_with_container(wait_codes)
        if command:
            logger.info(command)
            message = command + END_OF_LINE
            await self.stream.write(message.encode(encoding=self.encoding))
        if expected_codes or wait_codes:
            code, info = await self.parse_response()
            while any(map(code.matches, wait_codes)):
                code, info = await self.parse_response()
            if expected_codes:
                self.check_codes(expected_codes, code, info)
            return code, info

    @staticmethod
    def parse_epsv_response(s):
        """
        Parsing `EPSV` (`message (|||port|)`) response.

        :param s: response line
        :type s: :py:class:`str`

        :return: (ip, port)
        :rtype: (:py:class:`None`, :py:class:`int`)
        """
        matches = tuple(re.finditer(r"\((.)\1\1\d+\1\)", s))
        s = matches[-1].group()
        port = int(s[4:-2])
        return None, port

    @staticmethod
    def parse_pasv_response(s):
        """
        Parsing `PASV` server response.

        :param s: response line
        :type s: :py:class:`str`

        :return: (ip, port)
        :rtype: (:py:class:`str`, :py:class:`int`)
        """
        sub, *_ = re.findall(r"[^(]*\(([^)]*)", s)
        nums = tuple(map(int, sub.split(",")))
        ip = ".".join(map(str, nums[:4]))
        port = (nums[4] << 8) | nums[5]
        return ip, port

    @staticmethod
    def parse_directory_response(s):
        """
        Parsing directory server response.

        :param s: response line
        :type s: :py:class:`str`

        :rtype: :py:class:`pathlib.PurePosixPath`
        """
        seq_quotes = 0
        start = False
        directory = ""
        for ch in s:
            if not start:
                if ch == "\"":
                    start = True
            else:
                if ch == "\"":
                    seq_quotes += 1
                else:
                    if seq_quotes == 1:
                        break
                    elif seq_quotes == 2:
                        seq_quotes = 0
                        directory += '"'
                    directory += ch
        return pathlib.PurePosixPath(directory)

    @staticmethod
    def parse_unix_mode(s):
        """
        Parsing unix mode strings ("rwxr-x--t") into hexacimal notation.

        :param s: mode string
        :type s: :py:class:`str`

        :return mode:
        :rtype: :py:class:`int`
        """
        parse_rw = {"rw": 6, "r-": 4, "-w": 2, "--": 0}
        mode = 0
        mode |= parse_rw[s[0:2]] << 6
        mode |= parse_rw[s[3:5]] << 3
        mode |= parse_rw[s[6:8]]
        if s[2] == "s":
            mode |= 0o4100
        elif s[2] == "x":
            mode |= 0o0100
        elif s[2] != "-":
            raise ValueError

        if s[5] == "s":
            mode |= 0o2010
        elif s[5] == "x":
            mode |= 0o0010
        elif s[5] != "-":
            raise ValueError

        if s[8] == "t":
            mode |= 0o1000
        elif s[8] == "x":
            mode |= 0o0001
        elif s[8] != "-":
            raise ValueError

        return mode

    @staticmethod
    def format_date_time(d):
        """
        Formats dates from strptime in a consistent format

        :param d: return value from strptime
        :type d: :py:class:`datetime`

        :rtype: :py:class`str`
        """
        return d.strftime("%Y%m%d%H%M00")

    def parse_ls_date(self, s, *, now=None):
        """
        Parsing dates from the ls unix utility. For example,
        "Nov 18  1958" and "Nov 18 12:29".

        :param s: ls date
        :type s: :py:class:`str`

        :rtype: :py:class:`str`
        """
        with setlocale("C"):
            try:
                if now is None:
                    now = datetime.datetime.now()
                d = datetime.datetime.strptime(s, "%b %d %H:%M")
                d = d.replace(year=now.year)
                diff = (now - d).total_seconds()
                if diff > HALF_OF_YEAR_IN_SECONDS:
                    d = d.replace(year=now.year + 1)
                elif diff < -HALF_OF_YEAR_IN_SECONDS:
                    d = d.replace(year=now.year - 1)
            except ValueError:
                d = datetime.datetime.strptime(s, "%b %d  %Y")
        return self.format_date_time(d)

    def parse_list_line_unix(self, b):
        """
        Attempt to parse a LIST line (similar to unix ls utility).

        :param b: response line
        :type b: :py:class:`bytes` or :py:class:`str`

        :return: (path, info)
        :rtype: (:py:class:`pathlib.PurePosixPath`, :py:class:`dict`)
        """
        s = b.decode(encoding=self.encoding).rstrip()
        info = {}
        if s[0] == "-":
            info["type"] = "file"
        elif s[0] == "d":
            info["type"] = "dir"
        elif s[0] == "l":
            info["type"] = "link"
        else:
            info["type"] = "unknown"

        # TODO: handle symlinks(beware the symlink loop)
        info["unix.mode"] = self.parse_unix_mode(s[1:10])
        s = s[10:].lstrip()
        i = s.index(" ")
        info["unix.links"] = s[:i]

        if not info["unix.links"].isdigit():
            raise ValueError

        s = s[i:].lstrip()
        i = s.index(" ")
        info["unix.owner"] = s[:i]
        s = s[i:].lstrip()
        i = s.index(" ")
        info["unix.group"] = s[:i]
        s = s[i:].lstrip()
        i = s.index(" ")
        info["size"] = s[:i]

        if not info["size"].isdigit():
            raise ValueError

        s = s[i:].lstrip()
        info["modify"] = self.parse_ls_date(s[:12])
        s = s[12:].strip()
        if info["type"] == "link":
            i = s.rindex(" -> ")
            link_dst = s[i + 4:]
            link_src = s[:i]
            i = -2 if link_dst[-1] == "\'" or link_dst[-1] == "\"" else -1
            info["type"] = "dir" if link_dst[i] == "/" else "file"
            s = link_src
        return pathlib.PurePosixPath(s), info

    def parse_list_line_windows(self, b):
        """
        Parsing Microsoft Windows `dir` output

        :param b: response line
        :type b: :py:class:`bytes` or :py:class:`str`

        :return: (path, info)
        :rtype: (:py:class:`pathlib.PurePosixPath`, :py:class:`dict`)
        """
        line = b.decode(encoding=self.encoding).rstrip("\r\n")
        date_time_end = line.index("M")
        date_time_str = line[:date_time_end + 1].strip().split(" ")
        date_time_str = " ".join([x for x in date_time_str if len(x) > 0])
        line = line[date_time_end + 1:].lstrip()
        with setlocale("C"):
            strptime = datetime.datetime.strptime
            date_time = strptime(date_time_str, "%m/%d/%Y %I:%M %p")
        info = {}
        info["modify"] = self.format_date_time(date_time)
        next_space = line.index(" ")
        if line.startswith("<DIR>"):
            info["type"] = "dir"
        else:
            info["type"] = "file"
            info["size"] = line[:next_space].replace(",", "")
            if not info["size"].isdigit():
                raise ValueError
        # This here could cause a problem if a filename started with
        # whitespace, but if we were to try to detect such a condition
        # we would have to make strong assumptions about the input format
        filename = line[next_space:].lstrip()
        if filename == "." or filename == "..":
            raise ValueError
        return pathlib.PurePosixPath(filename), info

    def parse_list_line(self, b):
        """
        Parse LIST response with both Microsoft Windows® parser and
        UNIX parser

        :param b: response line
        :type b: :py:class:`bytes` or :py:class:`str`

        :return: (path, info)
        :rtype: (:py:class:`pathlib.PurePosixPath`, :py:class:`dict`)
        """
        ex = []
        parsers = (
            self.parse_list_line_custom,
            self.parse_list_line_unix,
            self.parse_list_line_windows,
        )
        for parser in parsers:
            if parser is None:
                continue
            try:
                return parser(b)
            except (ValueError, KeyError, IndexError) as e:
                ex.append(e)
        raise ValueError("All parsers failed to parse", b, ex)

    def parse_mlsx_line(self, b):
        """
        Parsing MLS(T|D) response.

        :param b: response line
        :type b: :py:class:`bytes` or :py:class:`str`

        :return: (path, info)
        :rtype: (:py:class:`pathlib.PurePosixPath`, :py:class:`dict`)
        """
        if isinstance(b, bytes):
            s = b.decode(encoding=self.encoding)
        else:
            s = b
        line = s.rstrip()
        facts_found, _, name = line.partition(" ")
        entry = {}
        for fact in facts_found[:-1].split(";"):
            key, _, value = fact.partition("=")
            entry[key.lower()] = value
        return pathlib.PurePosixPath(name), entry


class Client(BaseClient):
    """
    FTP client.

    :param socket_timeout: timeout for read operations
    :type socket_timeout: :py:class:`float`, :py:class:`int` or `None`

    :param read_speed_limit: download speed limit in bytes per second
    :type server_to_client_speed_limit: :py:class:`int` or `None`

    :param write_speed_limit: upload speed limit in bytes per second
    :type client_to_server_speed_limit: :py:class:`int` or `None`

    :param path_timeout: timeout for path-related operations (make directory,
        unlink file, etc.)
    :type path_timeout: :py:class:`float`, :py:class:`int` or
        :py:class:`None`

    :param path_io_factory: factory of «path abstract layer»
    :type path_io_factory: :py:class:`aioftp.AbstractPathIO`

    :param encoding: encoding to use for convertion strings to bytes
    :type encoding: :py:class:`str`

    :param ssl: if given and not false, a SSL/TLS transport is created
        (by default a plain TCP transport is created).
        If ssl is a ssl.SSLContext object, this context is used to create
        the transport; if ssl is True, a default context returned from
        ssl.create_default_context() is used.
        Please look :py:meth:`asyncio.loop.create_connection` docs.
    :type ssl: :py:class:`bool` or :py:class:`ssl.SSLContext`

    :param parse_list_line_custom: callable, which receive exactly one
        argument: line of type bytes. Should return tuple of Path object and
        dictionary with fields "modify", "type", "type", "size". For more
        information see sources.
    :type parse_list_line_custom: callable
    :param **siosocks_asyncio_kwargs: siosocks key-word only arguments
    """
    async def connect(self, host, port=DEFAULT_PORT):
        """
        :py:func:`asyncio.coroutine`

        Connect to server.

        :param host: host name for connection
        :type host: :py:class:`str`

        :param port: port number for connection
        :type port: :py:class:`int`
        """
        await super().connect(host, port)
        code, info = await self.command(None, "220", "120")
        return info

    async def login(self, user=DEFAULT_USER, password=DEFAULT_PASSWORD,
                    account=DEFAULT_ACCOUNT):
        """
        :py:func:`asyncio.coroutine`

        Server authentication.

        :param user: username
        :type user: :py:class:`str`

        :param password: password
        :type password: :py:class:`str`

        :param account: account (almost always blank)
        :type account: :py:class:`str`

        :raises aioftp.StatusCodeError: if unknown code received
        """
        code, info = await self.command("USER " + user, ("230", "33x"))
        while code.matches("33x"):
            if code == "331":
                cmd = "PASS " + password
            elif code == "332":
                cmd = "ACCT " + account
            else:
                raise errors.StatusCodeError("33x", code, info)
            code, info = await self.command(cmd, ("230", "33x"))

    async def get_current_directory(self):
        """
        :py:func:`asyncio.coroutine`

        Getting current working directory.

        :rtype: :py:class:`pathlib.PurePosixPath`
        """
        code, info = await self.command("PWD", "257")
        directory = self.parse_directory_response(info[-1])
        return directory

    async def change_directory(self, path=".."):
        """
        :py:func:`asyncio.coroutine`

        Change current directory. Goes «up» if no parameters passed.

        :param path: new directory, goes «up» if omitted
        :type path: :py:class:`str` or :py:class:`pathlib.PurePosixPath`
        """
        path = pathlib.PurePosixPath(path)
        if path == pathlib.PurePosixPath(".."):
            cmd = "CDUP"
        else:
            cmd = "CWD " + str(path)
        await self.command(cmd, "2xx")

    async def make_directory(self, path, *, parents=True):
        """
        :py:func:`asyncio.coroutine`

        Make directory.

        :param path: path to directory to create
        :type path: :py:class:`str` or :py:class:`pathlib.PurePosixPath`

        :param parents: create parents if does not exists
        :type parents: :py:class:`bool`
        """
        path = pathlib.PurePosixPath(path)
        need_create = []
        while path.name and not await self.exists(path):
            need_create.append(path)
            path = path.parent
            if not parents:
                break
        need_create.reverse()
        for path in need_create:
            await self.command("MKD " + str(path), "257")

    async def remove_directory(self, path):
        """
        :py:func:`asyncio.coroutine`

        Low level remove method for removing empty directory.

        :param path: empty directory to remove
        :type path: :py:class:`str` or :py:class:`pathlib.PurePosixPath`
        """
        await self.command("RMD " + str(path), "250")

    def list(self, path="", *, recursive=False, raw_command=None):
        """
        :py:func:`asyncio.coroutine`

        List all files and directories in "path".

        :param path: directory or file path
        :type path: :py:class:`str` or :py:class:`pathlib.PurePosixPath`

        :param recursive: list recursively
        :type recursive: :py:class:`bool`

        :param raw_command: optional ftp command to use in place of
            fallback logic (must be one of "MLSD", "LIST")
        :type raw_command: :py:class:`str`

        :rtype: :py:class:`list` or `async for` context

        ::

            >>> # lazy list
            >>> async for path, info in client.list():
            ...     # no interaction with client should be here(!)

            >>> # eager list
            >>> for path, info in (await client.list()):
            ...     # interaction with client allowed, since all paths are
            ...     # collected already

        ::

            >>> stats = await client.list()
        """
        class AsyncLister(AsyncListerMixin):
            stream = None

            async def _new_stream(cls, local_path):
                cls.path = local_path
                cls.parse_line = self.parse_mlsx_line
                if raw_command not in [None, "MLSD", "LIST"]:
                    raise ValueError("raw_command must be one of MLSD or "
                                     f"LIST, but got {raw_command}")
                if raw_command in [None, "MLSD"]:
                    try:
                        command = ("MLSD " + str(cls.path)).strip()
                        return await self.get_stream(command, "1xx")
                    except errors.StatusCodeError as e:
                        code = e.received_codes[-1]
                        if not code.matches("50x") or raw_command is not None:
                            raise
                if raw_command in [None, "LIST"]:
                    cls.parse_line = self.parse_list_line
                    command = ("LIST " + str(cls.path)).strip()
                    return await self.get_stream(command, "1xx")

            def __aiter__(cls):
                cls.directories = collections.deque()
                return cls

            async def __anext__(cls):
                if cls.stream is None:
                    cls.stream = await cls._new_stream(path)
                while True:
                    line = await cls.stream.readline()
                    while not line:
                        await cls.stream.finish()
                        if cls.directories:
                            current_path, info = cls.directories.popleft()
                            cls.stream = await cls._new_stream(current_path)
                            line = await cls.stream.readline()
                        else:
                            raise StopAsyncIteration

                    try:
                        name, info = cls.parse_line(line)
                    except Exception:
                        continue

                    stat = cls.path / name, info
                    if info["type"] == "dir" and recursive:
                        cls.directories.append(stat)
                    return stat

        return AsyncLister()

    async def stat(self, path):
        """
        :py:func:`asyncio.coroutine`

        Getting path stats.

        :param path: path for getting info
        :type path: :py:class:`str` or :py:class:`pathlib.PurePosixPath`

        :return: path info
        :rtype: :py:class:`dict`
        """
        path = pathlib.PurePosixPath(path)
        try:
            code, info = await self.command("MLST " + str(path), "2xx")
            name, info = self.parse_mlsx_line(info[1].lstrip())
            return info
        except errors.StatusCodeError as e:
            if not e.received_codes[-1].matches("50x"):
                raise

        for p, info in await self.list(path.parent):
            if p.name == path.name:
                return info
        else:
            raise errors.StatusCodeError(
                Code("2xx"),
                Code("550"),
                "path does not exists",
            )

    async def is_file(self, path):
        """
        :py:func:`asyncio.coroutine`

        Checks if path is file.

        :param path: path to check
        :type path: :py:class:`str` or :py:class:`pathlib.PurePosixPath`

        :rtype: :py:class:`bool`
        """
        info = await self.stat(path)
        return info["type"] == "file"

    async def is_dir(self, path):
        """
        :py:func:`asyncio.coroutine`

        Checks if path is dir.

        :param path: path to check
        :type path: :py:class:`str` or :py:class:`pathlib.PurePosixPath`

        :rtype: :py:class:`bool`
        """
        info = await self.stat(path)
        return info["type"] == "dir"

    async def exists(self, path):
        """
        :py:func:`asyncio.coroutine`

        Check path for existence.

        :param path: path to check
        :type path: :py:class:`str` or :py:class:`pathlib.PurePosixPath`

        :rtype: :py:class:`bool`
        """
        try:
            await self.stat(path)
            return True
        except errors.StatusCodeError as e:
            if e.received_codes[-1].matches("550"):
                return False
            raise

    async def rename(self, source, destination):
        """
        :py:func:`asyncio.coroutine`

        Rename (move) file or directory.

        :param source: path to rename
        :type source: :py:class:`str` or :py:class:`pathlib.PurePosixPath`

        :param destination: path new name
        :type destination: :py:class:`str` or :py:class:`pathlib.PurePosixPath`
        """
        await self.command("RNFR " + str(source), "350")
        await self.command("RNTO " + str(destination), "2xx")

    async def remove_file(self, path):
        """
        :py:func:`asyncio.coroutine`

        Low level remove method for removing file.

        :param path: file to remove
        :type path: :py:class:`str` or :py:class:`pathlib.PurePosixPath`
        """
        await self.command("DELE " + str(path), "2xx")

    async def remove(self, path):
        """
        :py:func:`asyncio.coroutine`

        High level remove method for removing path recursively (file or
        directory).

        :param path: path to remove
        :type path: :py:class:`str` or :py:class:`pathlib.PurePosixPath`
        """
        if await self.exists(path):
            info = await self.stat(path)
            if info["type"] == "file":
                await self.remove_file(path)
            elif info["type"] == "dir":
                for name, info in (await self.list(path)):
                    if info["type"] in ("dir", "file"):
                        await self.remove(name)
                await self.remove_directory(path)

    def upload_stream(self, destination, *, offset=0):
        """
        Create stream for write data to `destination` file.

        :param destination: destination path of file on server side
        :type destination: :py:class:`str` or :py:class:`pathlib.PurePosixPath`

        :param offset: byte offset for stream start position
        :type offset: :py:class:`int`

        :rtype: :py:class:`aioftp.DataConnectionThrottleStreamIO`
        """
        return self.get_stream(
            "STOR " + str(destination),
            "1xx",
            offset=offset,
        )

    def append_stream(self, destination, *, offset=0):
        """
        Create stream for append (write) data to `destination` file.

        :param destination: destination path of file on server side
        :type destination: :py:class:`str` or :py:class:`pathlib.PurePosixPath`

        :param offset: byte offset for stream start position
        :type offset: :py:class:`int`

        :rtype: :py:class:`aioftp.DataConnectionThrottleStreamIO`
        """
        return self.get_stream(
            "APPE " + str(destination),
            "1xx",
            offset=offset,
        )

    async def upload(self, source, destination="", *, write_into=False,
                     block_size=DEFAULT_BLOCK_SIZE):
        """
        :py:func:`asyncio.coroutine`

        High level upload method for uploading files and directories
        recursively from file system.

        :param source: source path of file or directory on client side
        :type source: :py:class:`str` or :py:class:`pathlib.Path`

        :param destination: destination path of file or directory on server
            side
        :type destination: :py:class:`str` or :py:class:`pathlib.PurePosixPath`

        :param write_into: write source into destination (if you want upload
            file and change it name, as well with directories)
        :type write_into: :py:class:`bool`

        :param block_size: block size for transaction
        :type block_size: :py:class:`int`
        """
        source = pathlib.Path(source)
        destination = pathlib.PurePosixPath(destination)
        if not write_into:
            destination = destination / source.name
        if await self.path_io.is_file(source):
            await self.make_directory(destination.parent)
            async with self.path_io.open(source, mode="rb") as file_in, \
                    self.upload_stream(destination) as stream:
                async for block in file_in.iter_by_block(block_size):
                    await stream.write(block)
        elif await self.path_io.is_dir(source):
            await self.make_directory(destination)
            sources = collections.deque([source])
            while sources:
                src = sources.popleft()
                async for path in self.path_io.list(src):
                    if write_into:
                        relative = destination.name / path.relative_to(source)
                    else:
                        relative = path.relative_to(source.parent)
                    if await self.path_io.is_dir(path):
                        await self.make_directory(relative)
                        sources.append(path)
                    else:
                        await self.upload(
                            path,
                            relative,
                            write_into=True,
                            block_size=block_size
                        )

    def download_stream(self, source, *, offset=0):
        """
        :py:func:`asyncio.coroutine`

        Create stream for read data from `source` file.

        :param source: source path of file on server side
        :type source: :py:class:`str` or :py:class:`pathlib.PurePosixPath`

        :param offset: byte offset for stream start position
        :type offset: :py:class:`int`

        :rtype: :py:class:`aioftp.DataConnectionThrottleStreamIO`
        """
        return self.get_stream("RETR " + str(source), "1xx", offset=offset)

    async def download(self, source, destination="", *, write_into=False,
                       block_size=DEFAULT_BLOCK_SIZE):
        """
        :py:func:`asyncio.coroutine`

        High level download method for downloading files and directories
        recursively and save them to the file system.

        :param source: source path of file or directory on server side
        :type source: :py:class:`str` or :py:class:`pathlib.PurePosixPath`

        :param destination: destination path of file or directory on client
            side
        :type destination: :py:class:`str` or :py:class:`pathlib.Path`

        :param write_into: write source into destination (if you want download
            file and change it name, as well with directories)
        :type write_into: :py:class:`bool`

        :param block_size: block size for transaction
        :type block_size: :py:class:`int`
        """
        source = pathlib.PurePosixPath(source)
        destination = pathlib.Path(destination)
        if not write_into:
            destination = destination / source.name
        if await self.is_file(source):
            await self.path_io.mkdir(destination.parent,
                                     parents=True, exist_ok=True)
            async with self.path_io.open(destination, mode="wb") as file_out, \
                    self.download_stream(source) as stream:
                async for block in stream.iter_by_block(block_size):
                    await file_out.write(block)
        elif await self.is_dir(source):
            await self.path_io.mkdir(destination, parents=True, exist_ok=True)
            for name, info in (await self.list(source)):
                full = destination / name.relative_to(source)
                if info["type"] in ("file", "dir"):
                    await self.download(name, full, write_into=True,
                                        block_size=block_size)

    async def quit(self):
        """
        :py:func:`asyncio.coroutine`

        Send "QUIT" and close connection.
        """
        await self.command("QUIT", "2xx")
        self.close()

    async def _do_epsv(self):
        code, info = await self.command("EPSV", "229")
        ip, port = self.parse_epsv_response(info[-1])
        return ip, port

    async def _do_pasv(self):
        code, info = await self.command("PASV", "227")
        ip, port = self.parse_pasv_response(info[-1])
        return ip, port

    async def get_passive_connection(self, conn_type="I",
                                     commands=("epsv", "pasv")):
        """
        :py:func:`asyncio.coroutine`

        Getting pair of reader, writer for passive connection with server.

        :param conn_type: connection type ("I", "A", "E", "L")
        :type conn_type: :py:class:`str`

        :param commands: sequence of commands to try to initiate passive
            server creation. First success wins. Default is EPSV, then PASV.
        :type commands: :py:class:`list`

        :rtype: (:py:class:`asyncio.StreamReader`,
            :py:class:`asyncio.StreamWriter`)
        """
        functions = {
            "epsv": self._do_epsv,
            "pasv": self._do_pasv,
        }
        if not commands:
            raise ValueError("No passive commands provided")
        await self.command("TYPE " + conn_type, "200")
        for i, name in enumerate(commands, start=1):
            name = name.lower()
            if name not in functions:
                raise ValueError(f"{name!r} not in {set(functions)!r}")
            try:
                ip, port = await functions[name]()
                break
            except errors.StatusCodeError as e:
                is_last = i == len(commands)
                if is_last or not e.received_codes[-1].matches("50x"):
                    raise
        if ip in ("0.0.0.0", None):
            ip = self.server_host
        reader, writer = await self._open_connection(ip, port)
        return reader, writer

    @async_enterable
    async def get_stream(self, *command_args, conn_type="I", offset=0):
        """
        :py:func:`asyncio.coroutine`

        Create :py:class:`aioftp.DataConnectionThrottleStreamIO` for straight
        read/write io.

        :param command_args: arguments for :py:meth:`aioftp.Client.command`

        :param conn_type: connection type ("I", "A", "E", "L")
        :type conn_type: :py:class:`str`

        :param offset: byte offset for stream start position
        :type offset: :py:class:`int`

        :rtype: :py:class:`aioftp.DataConnectionThrottleStreamIO`
        """
        reader, writer = await self.get_passive_connection(conn_type)
        if offset:
            await self.command("REST " + str(offset), "350")
        await self.command(*command_args)
        stream = DataConnectionThrottleStreamIO(
            self,
            reader,
            writer,
            throttles={"_": self.throttle},
            timeout=self.socket_timeout,
        )
        return stream

    async def abort(self, *, wait=True):
        """
        :py:func:`asyncio.coroutine`

        Request data transfer abort.

        :param wait: wait for abort response [426]→226 if `True`
        :type wait: :py:class:`bool`
        """
        if wait:
            await self.command("ABOR", "226", "426")
        else:
            await self.command("ABOR")


class ClientSession:
    """
    :py:class:`aioftp.Client` wrapper to use client as context manager. This
    make async call to :py:meth:`aioftp.Client.connect`,
    :py:meth:`aioftp.Client.login` on enter and :py:meth:`aioftp.Client.quit`
    on exit.

    :param host: host name for connection
    :type host: :py:class:`str`

    :param port: port number for connection
    :type port: :py:class:`int`

    :param user: username
    :type user: :py:class:`str`

    :param password: password
    :type password: :py:class:`str`

    :param account: account (almost always blank)
    :type account: :py:class:`str`

    :param **kwargs: keyword arguments, which passed to
        :py:class:`aioftp.Client`

    ::

        >>> async with aioftp.ClientSession("127.0.0.1") as client:
        ...     # do
    """

    def __init__(self, host, port=DEFAULT_PORT, user=DEFAULT_USER,
                 password=DEFAULT_PASSWORD, account=DEFAULT_ACCOUNT, **kwargs):
        self.host = host
        self.port = port
        self.user = user
        self.password = password
        self.account = account
        self.kwargs = kwargs

    async def __aenter__(self):
        self.client = Client(**self.kwargs)
        try:
            await self.client.connect(self.host, self.port)
            await self.client.login(self.user, self.password, self.account)
        except Exception:
            self.client.close()
            raise
        return self.client

    async def __aexit__(self, *exc_info):
        await self.client.quit()



import pytest

import aioftp


@pytest.mark.asyncio
async def test_permission_denied(pair_factory, Server,
                                 expect_codes_in_exception):
    s = Server([
        aioftp.User(permissions=[aioftp.Permission(writable=False)])
    ])
    async with pair_factory(None, s) as pair:
        with expect_codes_in_exception("550"):
            await pair.client.make_directory("foo")


@pytest.mark.asyncio
async def test_permission_overriden(pair_factory, Server):
    s = Server([
        aioftp.User(
            permissions=[
                aioftp.Permission("/", writable=False),
                aioftp.Permission("/foo"),
            ]
        )
    ])
    async with pair_factory(None, s) as pair:
        await pair.client.make_directory("foo")
        await pair.client.remove_directory("foo")



import pytest


@pytest.mark.asyncio
async def test_socks_success(pair_factory, Client, socks):
    client = Client(socks_host=socks.host, socks_port=socks.port,
                    socks_version=5, username="foo", password="bar")
    async with pair_factory(client):
        pass


@pytest.mark.asyncio
async def test_socks_fail(pair_factory, Client, socks):
    client = Client(socks_host=socks.host, socks_port=socks.port,
                    socks_version=5, username="bar", password="bar")
    with pytest.raises(ValueError):
        async with pair_factory(client):
            pass



from pathlib import Path

import pytest

import aioftp


@pytest.mark.asyncio
async def test_server_side_exception(pair_factory):
    class CustomServer(aioftp.Server):
        async def custom(*args, **kwargs):
            raise RuntimeError("Test error")
    factory = pair_factory(server_factory=CustomServer, do_quit=False)
    async with factory as pair:
        with pytest.raises(ConnectionResetError):
            await pair.client.command("custom", "200")


@pytest.mark.asyncio
async def test_bad_type_value(pair_factory, expect_codes_in_exception):
    async with pair_factory() as pair:
        with expect_codes_in_exception("502"):
            await pair.client.command("type FOO", "200")


@pytest.mark.asyncio
async def test_pbsz(pair_factory):
    async with pair_factory() as pair:
        await pair.client.command("pbsz", "200")


@pytest.mark.asyncio
async def test_prot(pair_factory, expect_codes_in_exception):
    async with pair_factory() as pair:
        await pair.client.command("prot P", "200")
        with expect_codes_in_exception("502"):
            await pair.client.command("prot foo", "200")


@pytest.mark.asyncio
async def test_server_ipv6_pasv(pair_factory, expect_codes_in_exception):
    async with pair_factory(host="::1", do_quit=False) as pair:
        with expect_codes_in_exception("503"):
            await pair.client.get_passive_connection(commands=["pasv"])


@pytest.mark.asyncio
async def test_epsv_extra_arg(pair_factory, expect_codes_in_exception):
    async with pair_factory(do_quit=False) as pair:
        with expect_codes_in_exception("522"):
            await pair.client.command("epsv foo", "229")


@pytest.mark.asyncio
async def test_bad_server_path_io(pair_factory, Server,
                                  expect_codes_in_exception):
    class BadPathIO(aioftp.MemoryPathIO):
        async def is_file(*a, **kw):
            return False

        async def is_dir(*a, **kw):
            return False

    s = Server(path_io_factory=BadPathIO)
    async with pair_factory(None, s) as pair:
        pio = pair.server.path_io_factory()
        async with pio.open(Path("/foo"), "wb"):
            pass
        await pair.client.list()



import pathlib
import asyncio
import datetime
import itertools

import pytest

import aioftp


def test_parse_directory_response():
    s = 'foo "baz "" test nop" """""fdfs """'
    parsed = aioftp.Client.parse_directory_response(s)
    assert parsed == pathlib.PurePosixPath('baz " test nop')


def test_connection_del_future():
    loop = asyncio.new_event_loop()
    c = aioftp.Connection(loop=loop)
    c.foo = "bar"
    del c.future.foo


def test_connection_not_in_storage():
    loop = asyncio.new_event_loop()
    c = aioftp.Connection(loop=loop)
    with pytest.raises(AttributeError):
        getattr(c, "foo")


def test_available_connections_too_much_acquires():
    ac = aioftp.AvailableConnections(3)
    ac.acquire()
    ac.acquire()
    ac.acquire()
    with pytest.raises(ValueError):
        ac.acquire()


def test_available_connections_too_much_releases():
    ac = aioftp.AvailableConnections(3)
    ac.acquire()
    ac.release()
    with pytest.raises(ValueError):
        ac.release()


def test_parse_pasv_response():
    p = aioftp.Client.parse_pasv_response
    assert p("(192,168,1,0,1,0)") == ("192.168.1.0", 256)


def test_parse_epsv_response():
    p = aioftp.Client.parse_epsv_response
    assert p("some text (ha-ha) (|||665|) ((((666() (|fd667s).") == (None, 666)
    assert p("some text (ha-ha) (|||665|) (6666666).") == (None, 666)


def _c_locale_time(d, format="%b %d %H:%M"):
    with aioftp.common.setlocale("C"):
        return d.strftime(format)


def test_parse_list_datetime_not_older_than_6_month_format():
    def date_to_p(d):
        return d.strftime("%Y%m%d%H%M00")
    p = aioftp.Client.parse_ls_date
    dates = (
        datetime.datetime(year=2002, month=1, day=1),
        datetime.datetime(year=2002, month=12, day=31),
    )
    dt = datetime.timedelta(seconds=15778476 // 2)
    deltas = (datetime.timedelta(), dt, -dt)
    for now, delta in itertools.product(dates, deltas):
        d = now + delta
        assert p(aioftp.Client, _c_locale_time(d), now=now) == date_to_p(d)


def test_parse_list_datetime_older_than_6_month_format():
    def date_to_p(d):
        return d.strftime("%Y%m%d%H%M00")
    p = aioftp.Client.parse_ls_date
    dates = (
        datetime.datetime(year=2002, month=1, day=1),
        datetime.datetime(year=2002, month=12, day=31),
    )
    dt = datetime.timedelta(seconds=15778476, days=30)
    deltas = (dt, -dt)
    for now, delta in itertools.product(dates, deltas):
        d = now + delta
        if delta.total_seconds() > 0:
            expect = date_to_p(d.replace(year=d.year - 1))
        else:
            expect = date_to_p(d.replace(year=d.year + 1))
        assert p(aioftp.Client, _c_locale_time(d), now=now) == expect


def test_parse_list_datetime_short():
    def date_to_p(d):
        return d.strftime("%Y%m%d%H%M00")
    p = aioftp.Client.parse_ls_date
    dates = (
        datetime.datetime(year=2002, month=1, day=1),
        datetime.datetime(year=2002, month=12, day=31),
    )
    for d in dates:
        s = _c_locale_time(d, format="%b %d  %Y")
        assert p(aioftp.Client, s) == date_to_p(d)


def test_parse_list_line_unix():
    lines = {
        "file": [
            "-rw-rw-r--  1 poh  poh   6595 Feb 27 04:14 history.rst",
            "lrwxrwxrwx  1 poh  poh      6 Mar 23 05:46 link-tmp.py -> tmp.py",
        ],
        "dir": [
            "drw-rw-r--  1 poh  poh   6595 Feb 27 04:14 history.rst",
        ],
        "unknown": [
            "Erw-rw-r--  1 poh  poh   6595 Feb 27 04:14 history.rst",
        ]
    }
    p = aioftp.Client(encoding="utf-8").parse_list_line_unix
    for t, stack in lines.items():
        for line in stack:
            _, parsed = p(line.encode("utf-8"))
            assert parsed["type"] == t
    with pytest.raises(ValueError):
        p(b"-rw-rw-r--  1 poh  poh   6xx5 Feb 27 04:14 history.rst")
    with pytest.raises(ValueError):
        p(b"-rw-rw-r--  xx poh  poh   6595 Feb 27 04:14 history.rst")


@pytest.mark.parametrize("owner", ["s", "x", "-", "E"])
@pytest.mark.parametrize("group", ["s", "x", "-", "E"])
@pytest.mark.parametrize("others", ["t", "x", "-", "E"])
@pytest.mark.parametrize("read", ["r", "-"])
@pytest.mark.parametrize("write", ["w", "-"])
def test_parse_unix_mode(owner, group, others, read, write):
    s = f"{read}{write}{owner}{read}{write}{group}{read}{write}{others}"
    if "E" in {owner, group, others}:
        with pytest.raises(ValueError):
            aioftp.Client.parse_unix_mode(s)
    else:
        assert isinstance(aioftp.Client.parse_unix_mode(s), int)


def test_parse_list_line_failed():
    with pytest.raises(ValueError):
        aioftp.Client(encoding="utf-8").parse_list_line(b"what a hell?!")


def test_reprs_works():
    repr(aioftp.Throttle())
    repr(aioftp.Permission())
    repr(aioftp.User())


def test_throttle_reset():
    t = aioftp.Throttle(limit=1, reset_rate=1)
    t.append(b"-" * 3, 0)
    assert t._start == 0
    assert t._sum == 3
    t.append(b"-" * 3, 2)
    assert t._start == 2
    assert t._sum == 4


def test_permission_is_parent():
    p = aioftp.Permission("/foo/bar")
    assert p.is_parent(pathlib.PurePosixPath("/foo/bar/baz"))
    assert not p.is_parent(pathlib.PurePosixPath("/bar/baz"))


def test_server_mtime_build():
    now = datetime.datetime(year=2002, month=1, day=1).timestamp()
    past = datetime.datetime(year=2001, month=1, day=1).timestamp()
    b = aioftp.Server.build_list_mtime
    assert b(now, now) == "Jan  1 00:00"
    assert b(past, now) == "Jan  1  2001"



import pathlib

import pytest

import aioftp


@pytest.mark.asyncio
async def test_create_and_remove_directory(pair_factory):
    async with pair_factory() as pair:
        await pair.client.make_directory("bar")
        (path, stat), *_ = files = await pair.client.list()
        assert len(files) == 1
        assert path == pathlib.PurePosixPath("bar")
        assert stat["type"] == "dir"

        await pair.client.remove_directory("bar")
        assert await pair.server_paths_exists("bar") is False


@pytest.mark.asyncio
async def test_create_and_remove_directory_long(pair_factory):
    async with pair_factory() as pair:
        await pair.client.make_directory("bar/baz")
        assert await pair.server_paths_exists("bar", "bar/baz")
        await pair.client.remove_directory("bar/baz")
        await pair.client.remove_directory("bar")
        assert await pair.server_paths_exists("bar") is False


@pytest.mark.asyncio
async def test_create_directory_long_no_parents(pair_factory):
    async with pair_factory() as pair:
        await pair.client.make_directory("bar/baz", parents=False)
        await pair.client.remove_directory("bar/baz")
        await pair.client.remove_directory("bar")


@pytest.mark.asyncio
async def test_change_directory(pair_factory):
    async with pair_factory() as pair:
        await pair.client.make_directory("bar")
        await pair.client.change_directory("bar")
        cwd = await pair.client.get_current_directory()
        assert cwd == pathlib.PurePosixPath("/bar")
        await pair.client.change_directory()
        cwd = await pair.client.get_current_directory()
        assert cwd == pathlib.PurePosixPath("/")


@pytest.mark.asyncio
async def test_change_directory_not_exist(pair_factory,
                                          expect_codes_in_exception):
    async with pair_factory() as pair:
        with expect_codes_in_exception("550"):
            await pair.client.change_directory("bar")


@pytest.mark.asyncio
async def test_rename_empty_directory(pair_factory):
    async with pair_factory() as pair:
        await pair.client.make_directory("bar")
        assert await pair.server_paths_exists("bar")
        assert await pair.server_paths_exists("baz") is False
        await pair.client.rename("bar", "baz")
        assert await pair.server_paths_exists("bar") is False
        assert await pair.server_paths_exists("baz")


@pytest.mark.asyncio
async def test_rename_non_empty_directory(pair_factory):
    async with pair_factory() as pair:
        await pair.make_server_files("bar/foo.txt")
        assert await pair.server_paths_exists("bar/foo.txt", "bar")
        await pair.client.make_directory("hurr")
        await pair.client.rename("bar", "hurr/baz")
        assert await pair.server_paths_exists("hurr/baz/foo.txt")
        assert await pair.server_paths_exists("bar") is False


class FakeErrorPathIO(aioftp.MemoryPathIO):

    def list(self, path):

        class Lister(aioftp.AbstractAsyncLister):
            @aioftp.pathio.universal_exception
            async def __anext__(self):
                raise Exception("KERNEL PANIC")

        return Lister(timeout=self.timeout)


@pytest.mark.asyncio
async def test_exception_in_list(pair_factory, Server,
                                 expect_codes_in_exception):
    s = Server(path_io_factory=FakeErrorPathIO)
    async with pair_factory(None, s) as pair:
        with expect_codes_in_exception("451"):
            await pair.client.list()


@pytest.mark.asyncio
async def test_list_recursive(pair_factory):
    async with pair_factory() as pair:
        await pair.make_server_files("foo/bar", "foo/baz/baz")
        files = await pair.client.list(recursive=True)
        assert len(files) == 4



import contextlib

import pytest

import aioftp


@contextlib.contextmanager
def universal_exception_reason(*exc):
    try:
        yield
    except aioftp.PathIOError as e:
        type, instance, traceback = e.reason
        m = f"Expect one of {exc}, got {instance}"
        assert isinstance(instance, exc), m
    else:
        raise Exception(f"No excepton. Expect one of {exc}")


def test_has_state(path_io):
    assert hasattr(path_io, "state")
    path_io.state


@pytest.mark.asyncio
async def test_exists(path_io, temp_dir):
    assert await path_io.exists(temp_dir)
    assert not await path_io.exists(temp_dir / "foo")


@pytest.mark.asyncio
async def test_is_dir(path_io, temp_dir):
    assert await path_io.is_dir(temp_dir)
    p = temp_dir / "foo"
    async with path_io.open(p, mode="wb"):
        pass
    assert not await path_io.is_dir(p)


@pytest.mark.asyncio
async def test_is_file(path_io, temp_dir):
    p = temp_dir / "foo"
    assert not await path_io.is_file(temp_dir)
    async with path_io.open(p, mode="wb"):
        pass
    assert await path_io.is_file(p)


@pytest.mark.asyncio
async def test_mkdir(path_io, temp_dir):
    p = temp_dir / "foo"
    assert not await path_io.exists(p)
    await path_io.mkdir(p)
    assert await path_io.exists(p)


@pytest.mark.asyncio
async def test_rmdir(path_io, temp_dir):
    p = temp_dir / "foo"
    assert not await path_io.exists(p)
    await path_io.mkdir(p)
    assert await path_io.exists(p)
    await path_io.rmdir(p)
    assert not await path_io.exists(p)


@pytest.mark.asyncio
async def test_ulink(path_io, temp_dir):
    p = temp_dir / "foo"
    assert not await path_io.exists(p)
    async with path_io.open(p, mode="wb"):
        pass
    assert await path_io.exists(p)
    await path_io.unlink(p)
    assert not await path_io.exists(p)


@pytest.mark.asyncio
async def test_list(path_io, temp_dir):
    d, f = temp_dir / "dir", temp_dir / "file"
    await path_io.mkdir(d)
    async with path_io.open(f, mode="wb"):
        pass
    paths = await path_io.list(temp_dir)
    assert len(paths) == 2
    assert set(paths) == {d, f}
    paths = set()
    async for p in path_io.list(temp_dir):
        paths.add(p)
    assert set(paths) == {d, f}


@pytest.mark.asyncio
async def test_stat(path_io, temp_dir):
    stat = await path_io.stat(temp_dir)
    for a in ["st_size", "st_mtime", "st_ctime", "st_nlink", "st_mode"]:
        assert hasattr(stat, a)


@pytest.mark.asyncio
async def test_open_context(path_io, temp_dir):
    p = temp_dir / "context"
    async with path_io.open(p, mode="wb") as f:
        await f.write(b"foo")
    async with path_io.open(p, mode="rb") as f:
        assert await f.read() == b"foo"
    async with path_io.open(p, mode="ab") as f:
        await f.write(b"bar")
    async with path_io.open(p, mode="rb") as f:
        assert await f.read() == b"foobar"
    async with path_io.open(p, mode="wb") as f:
        await f.write(b"foo")
    async with path_io.open(p, mode="rb") as f:
        assert await f.read() == b"foo"
    async with path_io.open(p, mode="r+b") as f:
        assert await f.read(1) == b"f"
        await f.write(b"un")
    async with path_io.open(p, mode="rb") as f:
        assert await f.read() == b"fun"


@pytest.mark.asyncio
async def test_open_plain(path_io, temp_dir):
    p = temp_dir / "plain"
    f = await path_io.open(p, mode="wb")
    await f.write(b"foo")
    await f.close()
    f = await path_io.open(p, mode="rb")
    assert await f.read() == b"foo"
    await f.close()
    f = await path_io.open(p, mode="ab")
    await f.write(b"bar")
    await f.close()
    f = await path_io.open(p, mode="rb")
    assert await f.read() == b"foobar"
    await f.close()
    f = await path_io.open(p, mode="wb")
    await f.write(b"foo")
    await f.close()
    f = await path_io.open(p, mode="rb")
    assert await f.read() == b"foo"
    await f.close()
    f = await path_io.open(p, mode="r+b")
    assert await f.read(1) == b"f"
    await f.write(b"un")
    await f.close()
    f = await path_io.open(p, mode="rb")
    assert await f.read() == b"fun"
    await f.close()


@pytest.mark.asyncio
async def test_file_methods(path_io, temp_dir):
    p = temp_dir / "foo"
    async with path_io.open(p, mode="wb") as f:
        await f.write(b"foo")
        with universal_exception_reason(ValueError):
            await path_io.seek(f, 0)
        await f.seek(0)
        await f.write(b"bar")
    async with path_io.open(p, mode="rb") as f:
        assert await f.read() == b"bar"


@pytest.mark.asyncio
async def test_rename(path_io, temp_dir):
    old = temp_dir / "foo"
    new = temp_dir / "bar"
    await path_io.mkdir(old)
    assert await path_io.exists(old)
    assert not await path_io.exists(new)
    await path_io.rename(old, new)
    assert not await path_io.exists(old)
    assert await path_io.exists(new)


def test_repr_works(path_io, temp_dir):
    repr(path_io)


@pytest.mark.asyncio
async def test_path_over_file(path_io, temp_dir):
    f = temp_dir / "file"
    async with path_io.open(f, mode="wb"):
        pass
    assert not await path_io.exists(f / "dir")


@pytest.mark.asyncio
async def test_mkdir_over_file(path_io, temp_dir):
    f = temp_dir / "file"
    async with path_io.open(f, mode="wb"):
        pass
    with universal_exception_reason(FileExistsError):
        await path_io.mkdir(f)


@pytest.mark.asyncio
async def test_mkdir_no_parents(path_io, temp_dir):
    p = temp_dir / "foo" / "bar"
    with universal_exception_reason(FileNotFoundError):
        await path_io.mkdir(p)


@pytest.mark.asyncio
async def test_mkdir_parent_is_file(path_io, temp_dir):
    f = temp_dir / "foo"
    async with path_io.open(f, mode="wb"):
        pass
    with universal_exception_reason(NotADirectoryError):
        await path_io.mkdir(f / "bar")


@pytest.mark.asyncio
async def test_mkdir_parent_is_file_with_parents(path_io, temp_dir):
    f = temp_dir / "foo"
    async with path_io.open(f, mode="wb"):
        pass
    with universal_exception_reason(NotADirectoryError):
        await path_io.mkdir(f / "bar", parents=True)


@pytest.mark.asyncio
async def test_rmdir_not_exist(path_io, temp_dir):
    with universal_exception_reason(FileNotFoundError):
        await path_io.rmdir(temp_dir / "foo")


@pytest.mark.asyncio
async def test_rmdir_on_file(path_io, temp_dir):
    f = temp_dir / "foo"
    async with path_io.open(f, mode="wb"):
        pass
    with universal_exception_reason(NotADirectoryError):
        await path_io.rmdir(f)


@pytest.mark.asyncio
async def test_rmdir_not_empty(path_io, temp_dir):
    f = temp_dir / "foo"
    async with path_io.open(f, mode="wb"):
        pass
    with universal_exception_reason(OSError):
        await path_io.rmdir(temp_dir)


@pytest.mark.asyncio
async def test_unlink_not_exist(path_io, temp_dir):
    with universal_exception_reason(FileNotFoundError):
        await path_io.unlink(temp_dir / "foo")


@pytest.mark.asyncio
async def test_unlink_on_dir(path_io, temp_dir):
    with universal_exception_reason(IsADirectoryError):
        await path_io.unlink(temp_dir)


@pytest.mark.asyncio
async def test_list_not_exist(path_io, temp_dir):
    assert await path_io.list(temp_dir / "foo") == []


@pytest.mark.asyncio
async def test_stat_not_exist(path_io, temp_dir):
    with universal_exception_reason(FileNotFoundError):
        await path_io.stat(temp_dir / "foo")


@pytest.mark.asyncio
async def test_open_read_not_exist(path_io, temp_dir):
    with universal_exception_reason(FileNotFoundError):
        await path_io.open(temp_dir / "foo", mode="rb")


@pytest.mark.asyncio
async def test_open_write_unreachable(path_io, temp_dir):
    with universal_exception_reason(FileNotFoundError):
        await path_io.open(temp_dir / "foo" / "bar", mode="wb")


@pytest.mark.asyncio
async def test_open_write_directory(path_io, temp_dir):
    with universal_exception_reason(IsADirectoryError):
        await path_io.open(temp_dir, mode="wb")


@pytest.mark.asyncio
async def test_open_bad_mode(path_io, temp_dir):
    with universal_exception_reason(ValueError):
        await path_io.open(temp_dir, mode="bad")


@pytest.mark.asyncio
async def test_rename_source_or_dest_parent_not_exist(path_io, temp_dir):
    with universal_exception_reason(FileNotFoundError):
        await path_io.rename(temp_dir / "foo", temp_dir / "bar")
    with universal_exception_reason(FileNotFoundError):
        await path_io.rename(temp_dir, temp_dir / "foo" / "bar")


@pytest.mark.asyncio
async def test_rename_over_exists(path_io, temp_dir):
    source = temp_dir / "source"
    destination = temp_dir / "destination"
    await path_io.mkdir(source)
    await path_io.mkdir(destination)
    await path_io.rename(source, destination)



"""ftp client/server for asyncio"""
# flake8: noqa

from .client import *
from .server import *
from .errors import *
from .common import *
from .pathio import *

__version__ = "0.14.0"
version = tuple(map(int, __version__.split(".")))

__all__ = (
    client.__all__ +
    server.__all__ +
    errors.__all__ +
    common.__all__ +
    pathio.__all__ +
    ("version", "__version__")
)



import pytest


async def not_implemented(connection, rest):
    connection.response("502", ":P")
    return True


@pytest.mark.asyncio
async def test_client_fallback_to_pasv_at_list(pair_factory):
    async with pair_factory(host="127.0.0.1") as pair:
        pair.server.epsv = not_implemented
        await pair.client.list()


@pytest.mark.asyncio
async def test_client_fail_fallback_to_pasv_at_list(pair_factory,
                                                    expect_codes_in_exception):
    async with pair_factory(host="127.0.0.1") as pair:
        pair.server.epsv = not_implemented
        with expect_codes_in_exception("502"):
            await pair.client.get_passive_connection(commands=["epsv"])


@pytest.mark.asyncio
async def test_passive_no_choices(pair_factory):
    async with pair_factory() as pair:
        with pytest.raises(ValueError):
            await pair.client.get_passive_connection(commands=[])


@pytest.mark.asyncio
async def test_passive_bad_choices(pair_factory):
    async with pair_factory() as pair:
        pair.server.epsv = not_implemented
        with pytest.raises(ValueError):
            await pair.client.get_passive_connection(commands=["FOO"])


@pytest.mark.parametrize("method", [("pasv", "227"), ("epsv", "229")])
@pytest.mark.asyncio
async def test_passive_multicall(pair_factory, method):
    async with pair_factory(host="127.0.0.1") as pair:
        code, info = await pair.client.command(*method)
        assert "created" in info[0]
        code, info = await pair.client.command(*method)
        assert "exists" in info[0]


@pytest.mark.parametrize("method", ["pasv", "epsv"])
@pytest.mark.asyncio
async def test_passive_closed_on_recall(pair_factory, method):
    async with pair_factory(host="127.0.0.1") as pair:
        r, w = await pair.client.get_passive_connection(commands=[method])
        nr, nw = await pair.client.get_passive_connection(commands=[method])
        with pytest.raises((ConnectionResetError, BrokenPipeError)):
            while True:
                w.write(b"-")
                await w.drain()
